header {* Messages and Paths *}

theory MessagesAndPaths imports CAN 
begin


types Message = "nat \<times> nat \<times> nat \<times> Zone" (* messageId source dest zoneToBeCovered *)

abbreviation message:: " nat => nat => nat => Zone  => Message"  ("</ _|/ _,/ _,/ _>" [0, 0, 0] 70)
where "<m|s,d,Z> \<equiv>(m,s,d,Z)"


definition Message_zone:: "Message \<Rightarrow> Zone"
where "Message_zone \<equiv> \<lambda>(m,s,d,Z).Z" (* zone to be covered by the next messages *)
definition Message_dest:: "Message \<Rightarrow> nat"
where "Message_dest \<equiv> \<lambda>(m,s,d,Z).d"
definition Message_source:: "Message \<Rightarrow> nat"
where "Message_source \<equiv> \<lambda>(m,s,d,Z).s"

text{*MARK*}
definition valid_path:: "Message set \<Rightarrow>Message list \<Rightarrow> bool" where 
"valid_path msgs ML \<equiv>  ML\<noteq>[] \<and>(\<forall> i<length ML. ML!i\<in>msgs) \<and> 
           (\<forall> i<length ML - 1. Message_dest (ML!i) = Message_source (ML!(Suc i))) 
           \<and> distinct ML"
                         
definition valid_path_set:: "Message set \<Rightarrow> nat \<Rightarrow> nat\<Rightarrow> Message list set"
where "valid_path_set msgs S D \<equiv> {ML. valid_path msgs ML \<and>Message_source (ML!0) =S\<and> Message_dest (ML!(length ML - 1)) = D}"
(*set of valid paths between S and D*)

abbreviation destination:: "Message list \<Rightarrow> nat"
where "destination MsgL \<equiv> Message_dest (MsgL!(length MsgL - 1))"
abbreviation source:: "Message list \<Rightarrow> nat"
where "source MsgL \<equiv> Message_source (MsgL!0)"

text{*MARK*}
definition path_inside_zone::"CAN\<Rightarrow>Message set\<Rightarrow>Zone\<Rightarrow>Message list set" where
"path_inside_zone C msgs Z \<equiv> 
 {MsgL. valid_path msgs MsgL \<and> source MsgL\<in>NodesInZone C Z \<and>
   (\<forall> i<length (MsgL). Message_dest (MsgL!i)\<in> NodesInZone C Z)}"


lemma valid_path_single[simp]:" (valid_path msgs [M]) = (M\<in>msgs)"
by (auto simp: valid_path_def);


lemma valid_path_append1[simp]: " (valid_path msgs (MsgL@[M])) = ((MsgL=[]\<and>M\<in>msgs)\<or>(valid_path msgs MsgL \<and> M\<notin>set MsgL \<and> (MsgL\<noteq>[] \<longrightarrow>Message_source M = destination MsgL) \<and> M\<in> msgs))"

apply (case_tac MsgL)
apply (simp add: valid_path_def nth_append);
apply (auto simp: valid_path_def);
apply (drule_tac x=i in spec);
apply (erule impE);
apply force;
apply (case_tac i);
apply (simp add:  nth_append)+;
apply (rotate_tac 1,drule_tac x=i in spec)
apply (case_tac i);
apply (simp add:  nth_append)+;
apply (rotate_tac 1,drule_tac x="length list" in spec);
apply (case_tac "length list");
apply (simp add:  nth_append)+;
apply (case_tac "i<Suc (length list)")
apply (drule_tac x=i in spec);
apply (case_tac i);
apply (simp add:  nth_append)+;
apply (subgoal_tac "i=Suc (length list)");
apply (simp add:  nth_append)+;
apply auto;
apply (case_tac "i<Suc (length list)")
apply (rotate_tac 1,drule_tac x=i in spec);
apply (case_tac i);
apply (simp add:  nth_append)+;
apply (subgoal_tac "i=(length list)");
apply (case_tac "length list");
apply (simp add:  nth_append)+;
done;

text{*MARK*}
lemma valid_path_cons[simp]:
"(valid_path msgs (M#MsgL)) = 
      ((MsgL=[] \<and> M\<in>msgs) \<or>
       (valid_path msgs MsgL \<and> M\<notin>set MsgL \<and> 
         (MsgL\<noteq>[]\<longrightarrow>Message_dest M=source MsgL) \<and> M\<in>msgs))"
apply (case_tac MsgL)
apply (simp add: valid_path_def);
apply (auto simp: valid_path_def);
apply (case_tac i);
apply simp+;
apply (case_tac i);
apply simp+;
done;

lemma valid_path_take[rule_format]: " \<forall>i>0. valid_path msgs MsgL \<longrightarrow> valid_path msgs (take i MsgL)"
apply (rule rev_induct);
apply (auto);
apply (case_tac "i-length xs");
apply auto;
done;

lemma valid_path_drop[rule_format]: " \<forall>i<length MsgL. valid_path msgs MsgL \<longrightarrow> valid_path msgs (drop i MsgL)"
apply (induct_tac MsgL)
apply (auto)
apply (case_tac "i")
apply auto
done

lemma valid_in_valid_path_set[intro]:
"valid_path msgs MsgL = (MsgL\<in> (valid_path_set msgs (Message_source(MsgL!0)) (destination MsgL)))"
apply (auto simp: valid_path_set_def)
done;

lemma valid_in_valid_path_set_rev[intro]:
"MsgL\<in> (valid_path_set msgs S D) \<Longrightarrow>valid_path msgs MsgL"
apply (auto simp: valid_in_valid_path_set); apply (simp add: valid_path_set_def);
done;

lemma valid_in_msgs[intro]:
"\<lbrakk>valid_path msgs MsgL; i<length MsgL\<rbrakk>\<Longrightarrow>MsgL!i\<in>msgs"
by (auto simp: valid_path_def);


lemma finite_valid_path: "finite msgs \<Longrightarrow> finite {MsgL. valid_path msgs MsgL}"
apply (unfold valid_path_def);
apply auto;
apply (drule finite_distinct_list)
apply (subgoal_tac "{MsgL. MsgL \<noteq> [] \<and> (\<forall>i<length MsgL. MsgL ! i \<in> msgs) \<and> (\<forall>i<length MsgL - Suc 0. Message_dest (MsgL ! i) = Message_source (MsgL ! Suc i)) \<and> distinct MsgL}\<subseteq>{xs. distinct xs \<and> set xs \<subseteq> msgs}")
apply (auto simp: set_conv_nth);
apply (rule finite_subset);
apply auto;
done;

text{*MAKR*}
lemma finite_path_inside_zone:
"finite msgs \<Longrightarrow>finite (path_inside_zone C msgs Z)"
apply (unfold path_inside_zone_def)
apply (insert finite_valid_path[of msgs])
apply auto;
done;

subsection{* list_message_nodes: list of nodes covered by the zones of a list of messages *}


(* list_message_nodes returns the set of nodes intersecting the zones associated to a list of messages *)
primrec list_message_nodes:: "CAN \<Rightarrow>Message list \<Rightarrow> nat set"
where
empty_list_msg: "list_message_nodes C [] = {}" |
step_list_msg: "list_message_nodes C (M#ml) = NodesInZone C (Message_zone M) \<union>list_message_nodes C ml"

(*alternative definition*)
lemma list_message_nodes_set: 
"list_message_nodes C ml = \<Union> set (map (\<lambda>M. (NodesInZone C (Message_zone M))) ml)"
apply (induct_tac ml);
apply auto;
done;

(*alternative definition based on existence of ta nth member in message list*)
lemma list_message_nodes_nth[rule_format]: 
"list_message_nodes C ml = {N \<in> CAN_Nodes C. \<exists> x. ( (\<exists>i<length ml.(x\<in> Message_zone (ml!i)))\<and>x\<in> CAN_Zones C N)}"
apply (induct_tac ml);
 apply auto;
  apply (unfold intersects_def)
  apply clarify;
  apply (rule_tac x=tup in exI);
  apply simp+;
  apply (rule_tac x=0 in exI);
  apply force;
 apply (rule_tac x=xa in exI);
 apply force;
apply (case_tac i);
 apply auto;
done
(*simp rule to induct list_message_nodes on list*)
lemma list_message_nodes_cons[rule_format,simp]: " list_message_nodes C (ml@ml') =(list_message_nodes C ml \<union>list_message_nodes C ml')"
apply (induct_tac ml)
apply auto;
done;

(*a node in list_message_nodes is a node of the CAN*)
lemma list_message_zone_CAN_Node[intro,rule_format]: "N\<in> list_message_nodes C ml \<longrightarrow> N\<in> CAN_Nodes C"
apply (induct_tac ml);
apply auto;
done

subsection{*list_distinct_nodes: non-intersection of two zones of a message list*}
(* list_distinct_nodes takes a list of messages ml and returns true if there is no node intersected by the zones of 2 messages in ml*)
primrec list_distinct_nodes:: "CAN \<Rightarrow>Message list \<Rightarrow> bool"
where
"list_distinct_nodes C  [] = True" |
" list_distinct_nodes C (M#ml) = ((list_distinct_nodes C ml) \<and> (NodesInZone C (Message_zone M) \<inter> list_message_nodes C ml ={})) "

(*alternative definition*)
lemma list_distinct_nodes_def2: 
"list_distinct_nodes C ml = 
(\<forall> i<length ml. \<forall> j<length ml. NodesInZone C (Message_zone (ml!i)) \<inter> NodesInZone C(Message_zone (ml!j)) \<noteq>{} \<longrightarrow> i=j)"
apply (induct_tac ml)
 apply simp+;
apply clarify;
apply rule;
 apply (intro allI impI);
 apply auto;
   apply (case_tac i);
    apply (case_tac j);
     apply auto
    apply (simp add: list_message_nodes_nth intersects_def)
    apply force;
   apply (case_tac j);
    apply auto
(*4*)
   apply (simp add: intersects_def)
   apply clarify;
   apply (simp add: list_message_nodes_nth);
   apply force;
  apply (drule_tac x=0 in spec)
  apply simp
  apply (simp add: list_message_nodes_nth);
  apply clarify;
  apply (rotate_tac 2)
  apply (drule_tac x="Suc i" in spec)
  apply (simp add: intersects_def);
  apply force;
(*2*)
 apply (drule_tac x="Suc ia" in spec);
 apply simp;
 apply (drule_tac x="Suc ja" in spec);
 apply force;
apply (drule_tac x="Suc i" in spec);
apply simp;
apply (drule_tac x="Suc j" in spec);
apply force;
done

lemma list_distinct_nodes_cons[rule_format,simp]: " (list_distinct_nodes C (ml@ml') =(list_distinct_nodes C ml \<and>list_distinct_nodes C ml' \<and>list_message_nodes C ml\<inter>list_message_nodes C ml' = {}))"
apply (induct_tac ml)
apply auto;
done

end