header {* broadcast *}

theory broadcast imports MessagesAndPaths  Induct Connected
begin

text{*MARK*}
typedef Broadcast = "{(can,msgs,initiator). 
(\<forall> x y m Z m' Z'. (<m|x,y,Z>\<in>msgs \<and> <m'|x,y,Z'>\<in>msgs)\<longrightarrow>(m=m' \<and> Z=Z')) \<and>
  (initiator \<in> CAN_Nodes can)\<and>
  (\<forall> m s d Z. <m|s,d,Z>\<in>msgs \<longrightarrow> 
   (s\<in> CAN_Nodes can \<and> d\<in> CAN_Nodes can \<and> (d=initiator\<or>CAN_neighbour can s d) \<and>
    (s=initiator\<or>(\<exists> MsgL. valid_path msgs MsgL\<and>destination MsgL=s\<and>source MsgL=initiator))      
)) }"
apply auto;
apply (rule_tac x="Abs_CAN ({0},(\<lambda>n. Entire_Space),{})" in exI)
apply (unfold CAN_Nodes_def);
apply (rule_tac x="{}" in exI);
apply auto;
apply (rule_tac x="0" in exI);
apply (subgoal_tac "({0},(\<lambda>n. Entire_Space),{})\<in> CAN") 
apply (auto simp: CAN_def Abs_CAN_inverse  Nat_def Space_def Abs_Space_inverse Rep_Space)
apply (fold Nat_def)
apply (rule everything_smallerthan_max)
apply (rule_tac x="\<lambda>n.(Abs_Space 0)" in exI)
apply (simp add: Space_def Nat_def Abs_Space_inverse)
done


abbreviation broadcast:: " CAN => Message set => nat  => Broadcast"  ("<_,/ _,/ _>" [0, 0] 70)
where "<C,M,n> \<equiv> Abs_Broadcast (C,M,n)"

definition BC_msgs:: "Broadcast \<Rightarrow> Message set"
where "BC_msgs BC \<equiv> (\<lambda>(a,b,c).b) (Rep_Broadcast BC)"

definition BC_CAN:: "Broadcast \<Rightarrow> CAN"
where "BC_CAN BC \<equiv> (\<lambda>(a,b,c).a) (Rep_Broadcast BC)"

definition BC_initiator:: "Broadcast \<Rightarrow> nat"
where "BC_initiator BC \<equiv> (\<lambda>(a,b,c).c) (Rep_Broadcast BC)"

lemma BC_msgs_simp[simp]: "(C,M,n) \<in> Broadcast \<Longrightarrow>BC_msgs (<C,M,n>) = M"
apply (simp add: BC_msgs_def);
apply (drule Abs_Broadcast_inverse);
apply auto;
done;

definition Zoneij:: "Broadcast \<Rightarrow> (nat \<times> nat \<times> Zone) set"
where "Zoneij BC \<equiv> {(s,d,Z). \<exists> m. <m|s,d,Z>\<in>BC_msgs BC}"

lemma UniqueZoneij: "(s,d,Z)\<in> Zoneij BC \<and> (s,d,Z')\<in> Zoneij BC \<Longrightarrow> Z=Z'"
apply (unfold Zoneij_def BC_msgs_def);
apply simp+;
apply (subgoal_tac "Rep_Broadcast BC\<in> Broadcast")
 apply (simp add: Broadcast_def)
 apply auto;
  apply (drule_tac x=s in spec);
  apply (rotate_tac -1,drule_tac x=d in spec);
  apply (rotate_tac -1,drule_tac x=m in spec);
  apply (rotate_tac -1,drule_tac x=Z in spec);
  apply force;
(*2*)
 apply (drule_tac x=s in spec);
 apply (rotate_tac -1,drule_tac x=d in spec);
 apply (rotate_tac -1,drule_tac x=m in spec);
 apply (rotate_tac -1,drule_tac x=Z in spec);
 apply force;
apply (rotate_tac 1, drule sym);
apply (auto simp: Rep_Broadcast);
done;

text{*set of messages direct consequences (next step) of the given message*}
definition Consequences:: "Broadcast \<Rightarrow> Message \<Rightarrow> Message set" 
where "Consequences BC M = {M'\<in> BC_msgs BC. \<exists> m s d Z m'  d' Z'. M= <m|s,d,Z> \<and> M' = <m'|d,d',Z'>}"

definition BC_always_reduce_zones:: "Broadcast\<Rightarrow>bool"
-- "the set of zones to be covered by each of the neighboursconsequence  is the zone to be covered by the initial message minus the consequence nodes ' zones"
where "BC_always_reduce_zones BC\<equiv> 
\<forall> M \<in> (BC_msgs BC). \<Union> Message_zone ` (Consequences BC M) = 
Message_zone M - \<Union> (\<lambda>x. CAN_Zones (BC_CAN BC) (Message_dest x)) ` (Consequences BC M)"

(* UNUSED
definition Non_intersecting_zones::  "Broadcast\<Rightarrow>bool"
where "Non_intersecting_zones BC\<equiv>
\<forall> M \<in> (BC_msgs BC). \<forall> M'\<in> (Consequences BC M). \<forall> M''\<in> (Consequences BC M). 
                (M'=M''\<or> Message_zone M' \<inter> Message_zone M'' = {} )"
*)

definition set_of_valid_path_from_origin:: "Broadcast \<Rightarrow> nat \<Rightarrow> Message list set"
  -- "list of valid paths reaching (from initator)  a given node N*"
where "set_of_valid_path_from_origin BC N \<equiv> valid_path_set (BC_msgs BC) (BC_initiator BC) N"

text{*MARK*}
definition Coverage:: "Broadcast\<Rightarrow>bool" where 
"Coverage BC \<equiv> \<forall> n\<in>CAN_Nodes (BC_CAN BC). 
           (n=BC_initiator BC \<or> (\<exists> m s Z. <m|s,n,Z>\<in>BC_msgs BC))"
 -- "coverage = all nodes destination of a message (except initator)"

(*CHECK: for a zone of a message, for each node in the zone, there is a valid path arriving at the node (nothing enforces the path is inside the zone) *)
definition Zone_coverage:: "Broadcast \<Rightarrow> bool"
where "Zone_coverage BC \<equiv> \<forall> m s d Z. <m|s,d,Z> \<in> BC_msgs BC \<longrightarrow> 
           (\<forall> N\<in> (CAN_Nodes (BC_CAN BC)). intersects (CAN_Zones (BC_CAN BC) N) Z \<longrightarrow>
             (\<exists> MsgL. valid_path (BC_msgs BC) MsgL \<and> destination MsgL = N \<and> source MsgL=d))"

(* there is a valid path reaching the destination of each message (from the initiator) *)lemma Reachable_msgs_path[rule_format]: "\<lbrakk> <m|s,d,Z>\<in> BC_msgs BC;d\<noteq>BC_initiator BC\<rbrakk> \<Longrightarrow> (valid_path_set (BC_msgs BC) (BC_initiator BC) d \<noteq>{})"
apply (auto simp: valid_path_set_def)
apply (subgoal_tac "Rep_Broadcast BC \<in> Broadcast");
 apply ( simp add: Broadcast_def BC_msgs_def)
 apply auto;
 apply (rotate_tac -1, drule_tac x=m in spec)
 apply (rotate_tac -1, drule_tac x=s in spec)
 apply (rotate_tac -1, drule_tac x=d in spec)
 apply (erule impE, force)
apply (auto simp: BC_initiator_def)
apply (rule_tac x="[<m| s, d, Z>]" in exI)
apply (auto simp: nth_append valid_path_cons Message_dest_def Message_source_def )
apply (case_tac  "< m| Message_dest (MsgL ! (length MsgL - Suc 0)), d, Z> \<in> set MsgL");
apply (simp add: List.in_set_conv_nth)
apply clarify;
apply (rule_tac x="take (Suc i) MsgL" in exI)
apply auto;
apply (rule valid_path_take)
apply simp+;
apply (subgoal_tac "min (length MsgL) (Suc i) = (Suc i)");
apply (force simp: Message_dest_def);
apply simp;
apply (rule_tac x="MsgL@[< m| Message_dest (MsgL ! (length MsgL - Suc 0)), d, Z> ]" in exI)
apply (auto simp: nth_append valid_path_cons Message_dest_def Message_source_def );
apply (simp add: valid_path_def)
apply (rule Rep_Broadcast);
done

(*take a message < m|s,d,Z> in a path, any subsequent message necessarilly arrives in zone Z *)

lemma finite_msgs: "finite (BC_msgs BC)"
apply (subgoal_tac "Rep_Broadcast BC\<in>Broadcast")
 apply (simp add: BC_msgs_def Broadcast_def)
 apply auto;
 apply (insert finite_CAN_Nodes)
 apply (drule_tac x=x in meta_spec);
 apply (subgoal_tac " \<exists> f. (\<forall> m s d Z. ((< m| s, d, Z> \<in> a) \<longrightarrow> (m=f(s,d))))")
  apply (subgoal_tac " \<exists> f. (\<forall> m s d Z. ((< m| s, d, Z> \<in> a) \<longrightarrow> (Z=f(s,d))))")
   apply clarify;
   apply simp+;
(*4*)
   apply (subgoal_tac "a\<subseteq>{ < f(s,d)| s, d, fa(s,d)>| s d.  s \<in> CAN_Nodes x \<and>
                       d \<in> CAN_Nodes x}");
    apply (rule finite_subset);
     apply simp+;
    apply (subgoal_tac "finite ((CAN_Nodes x)\<times>(CAN_Nodes x))")
     apply (subgoal_tac "{< f (s, d)| s, d, fa (s, d)> |s d. s \<in> CAN_Nodes x \<and> d \<in> CAN_Nodes x} = 
                        (\<lambda> sd. < f (fst sd, snd sd)| fst sd, snd sd, fa (fst sd, snd sd)>) ` (CAN_Nodes x \<times> CAN_Nodes x)")
      apply (drule_tac h="\<lambda> sd. < f (fst sd, snd sd)| fst sd, snd sd, fa (fst sd, snd sd)>" in  Finite_Set.finite_imageI)
      apply simp;
     apply rule;
      apply clarify;
      apply (subgoal_tac "< f (s, d)| s, d, fa (s, d)> = (\<lambda>sd. < f (fst sd, snd sd)| fst sd, snd sd, fa (fst sd, snd sd)>) (s,d)")
(*8*)
       apply (rotate_tac -1, erule ssubst);
       apply (rule imageI);
       apply force;
      apply force;
     apply force;
    apply force;
(*4*)
   apply rule;
   apply clarify;
   apply (rotate_tac 2,drule_tac x=aa in spec)
   apply (rotate_tac -1,drule_tac x=ab in spec)
   apply (rotate_tac -1,drule_tac x=ac in spec)
   apply (erule impE, force)
   apply simp;
   apply (rotate_tac 2,drule_tac x=aa in spec)
   apply (rotate_tac -1,drule_tac x=ab in spec)
   apply (rotate_tac -1,drule_tac x=ac in spec)
   apply force;
  apply (rule_tac x="\<lambda> sd. (the_elem{Z. \<exists> m. < m| fst sd, snd sd, Z> \<in> a})" in exI);
  apply simp+;
  apply clarify;
  apply (subgoal_tac "{Z. \<exists>m. < m| s, d, Z> \<in> a}={Z}")
   apply force;
(*3*)
  apply (drule_tac x=s in spec)
  apply (rotate_tac -1,drule_tac x=d in spec)
  apply (rotate_tac -1,drule_tac x=m in spec)
  apply (rotate_tac -1,drule_tac x=Z in spec)
  apply rule;
   apply force;
  apply rule;
  apply clarify;
  apply (rotate_tac -1,drule_tac x=m in spec)
  apply (rotate_tac -1,drule_tac x=Z in spec)
  apply force;

(*2*)
 apply (rule_tac x="\<lambda> sd. (the_elem{m. \<exists> Z. < m| fst sd, snd sd, Z> \<in> a})" in exI);
 apply simp+;
 apply clarify;
 apply (subgoal_tac "{m. \<exists>Z. < m| s, d, Z> \<in> a}={m}")
  apply force;
 apply (drule_tac x=s in spec)
 apply (rotate_tac -1,drule_tac x=d in spec)
 apply (rotate_tac -1,drule_tac x=m in spec)
 apply (rotate_tac -1,drule_tac x=Z in spec)
 apply rule;
(*3*)
  apply force;
 apply rule;
 apply clarify;
 apply (rotate_tac -1,drule_tac x=m in spec)
 apply (rotate_tac -1,drule_tac x=Z in spec)
 apply force;
apply (rule Rep_Broadcast);
done;


lemma FiniteConsequences: "finite (Consequences BC M)"
apply (unfold Consequences_def BC_msgs_def)
apply (insert finite_msgs[of BC])
apply (auto simp: BC_msgs_def)
done

(*(if there is a message inside a zone) then there is a longest valid path inside the zone*)
lemma longest_path: "\<lbrakk> finite msgs; \<exists> M\<in> msgs. (Message_dest M)\<in>NodesInZone C Z \<and>(Message_source M)\<in>NodesInZone C Z\<rbrakk>
          \<Longrightarrow> \<exists> M\<in>(path_inside_zone C msgs Z). \<forall> M'\<in>(path_inside_zone C msgs Z). length M'\<le> length M"
apply (rule boundedSet);
apply (rule finite_path_inside_zone)
apply assumption;
apply clarify;
apply (subgoal_tac "[< a| aa, ab, b>]\<in> path_inside_zone C msgs Z");
apply force;
apply (unfold path_inside_zone_def);
apply rule;
apply force;
done

lemma longest_path_2: "\<lbrakk> finite msgs; path_inside_zone C msgs Z\<noteq>{} \<rbrakk>
          \<Longrightarrow> \<exists> M \<in> (path_inside_zone C msgs Z). \<forall> M'\<in>(path_inside_zone C msgs Z). length M'\<le> length M"
apply (rule boundedSet);
apply (rule finite_path_inside_zone)
apply assumption;
apply clarify;
done
lemma longest_path_BC: " \<exists> M\<in> (BC_msgs BC). (Message_dest M\<in>NodesInZone C Z \<and> (Message_source M)\<in>NodesInZone C Z)
          \<Longrightarrow> \<exists> M \<in> (path_inside_zone C (BC_msgs BC) Z). \<forall> M'\<in>(path_inside_zone C (BC_msgs BC) Z). length M'\<le> length M"
 -- "idem but with messages of a broadcast"
apply (insert longest_path [of "BC_msgs BC" C Z])
apply auto
apply (erule_tac meta_impE)
apply (simp add: finite_msgs)+
done




text{*MARK*}
definition Set_of_Valid_ZNL::"CAN \<Rightarrow>nat \<Rightarrow>Zone\<Rightarrow>((Zone\<times>nat) list) set" where
" Set_of_Valid_ZNL C N Z \<equiv> {ZNL.
 (\<forall> i<length ZNL. (snd (ZNL!i)\<in> CAN_Nodes C
   \<and> CAN_neighbour C N (snd (ZNL!i)) \<and> Connected C (fst (ZNL!i))
   \<and> (\<forall> j<length ZNL. j\<noteq>i\<longrightarrow>snd (ZNL!i)\<noteq>snd (ZNL!j)) 
   \<and> snd (ZNL!i)\<in>NodesInZone C (fst (ZNL!i)))) 
 \<and> (\<Union>i<length ZNL. (fst (ZNL!i))) = Z
}"
-- " A ZNL associates to a node (source) and a (source) zone of a given CAN, the list of (destination) nodes and zones the message must be forwarded to 
A ZNL is Valid iff:
  - each destination is a valid node, neighbour of the source
  - each zone is connected
  - each destination node belong to its zone
  - and the set of destination zones covers the source zone
"

definition Set_of_Optimal_ZNL::  "CAN \<Rightarrow>nat \<Rightarrow>Zone \<Rightarrow> ((Zone\<times>nat) list) set"
where
" Set_of_Optimal_ZNL C N Z \<equiv> {ZNL.
   (\<forall> i<length ZNL. 
          (snd(ZNL!i)\<in> CAN_Nodes C 
          \<and> CAN_neighbour C N (snd(ZNL!i))
          \<and> Connected C (fst(ZNL!i)) 
          \<and> snd(ZNL!i)\<in>NodesInZone C (fst(ZNL!i))
          \<and> (\<forall> j<length ZNL. j\<noteq>i\<longrightarrow>( (snd(ZNL!i))\<noteq> (snd(ZNL!j))
                        \<and> NoNodeInBothZones C (fst(ZNL!i)) (fst(ZNL!j)))) ))
  \<and> (\<Union>i<length ZNL. (fst(ZNL!i))) = Z
}"

-- " A  ZNL is Optimal iff:
Same as Valid PLUS:
  - no overlap: 2 zones of the same source are disjoint (no node in both zones)
"

(*optimal \<Rightarrow> valid*)
lemma Optimal_isValid_ZNL: "ZNL\<in>Set_of_Optimal_ZNL C N Z\<Longrightarrow>ZNL\<in>Set_of_Valid_ZNL C N Z"
apply (auto simp: Set_of_Optimal_ZNL_def Set_of_Valid_ZNL_def)
done

lemma nonEmptyZNL[simp]:" ([]\<in> Set_of_Valid_ZNL C N Z) = (Z={})"
-- "a ZNL can only be empty for an empty zone "
apply (unfold Set_of_Valid_ZNL_def)
apply auto;
done

lemma ConnectedZNL_Optimal: "Connected C Z \<Longrightarrow> \<forall>N'\<in>CAN_Nodes C. (CAN_neighbour C N N'\<and> intersects (CAN_Zones C N') Z)  \<longrightarrow> [(Z,N')]\<in>Set_of_Optimal_ZNL C N Z"
-- "if a zone is connected, then the single NL consisting of a neighbour of N and the entire zone is a Optimal ZNL, ie [(Z,N')] is a Optimal (ZNL C N Z)  2
NB: remember optimal \<Rightarrow> Valid "
apply (unfold Set_of_Optimal_ZNL_def)
apply auto;
done;

(* destination nodes of a valid znl are distinct *)
lemma  distinct_ZNL_Nodes: "ZNL\<in>Set_of_Valid_ZNL C N Z \<Longrightarrow> distinct (map snd ZNL)"
apply (simp add: Set_of_Valid_ZNL_def List.distinct_conv_nth);
done;

(* destination zones of a znl are distinct *) (* requires optimality *)
lemma  distinct_ZNL_Zones_Optimal: "ZNL\<in>Set_of_Optimal_ZNL C N Z \<Longrightarrow> distinct (map fst ZNL)"
apply (simp add: Set_of_Optimal_ZNL_def List.distinct_conv_nth);
apply auto;
apply (frule_tac x=i in spec);
apply (frule_tac x=j in spec);
apply (auto)
done;


(* combination of previous lemmas same destination zone \<Rightarrow> same destination node *) (* requires optimality *)
lemma distinc_ZNL_Nodes_set_Optimal: "\<lbrakk>ZNL\<in>Set_of_Optimal_ZNL C N Z; (Z,N)\<in>set ZNL;(Z,N')\<in> set ZNL\<rbrakk> \<Longrightarrow>N=N'"
apply (frule Optimal_isValid_ZNL)
apply (frule distinct_ZNL_Nodes,drule distinct_ZNL_Zones_Optimal);
apply (auto simp: distinct_conv_nth set_conv_nth);
apply (drule_tac x=i in spec,auto);
apply (drule_tac x=ia in spec,auto);
apply (drule sym, simp)
apply (drule sym, auto)
done;

(*destination nodes are different from source node*)
lemma ZNL_diff: "\<lbrakk>ZNL\<in> Set_of_Valid_ZNL C N Z; (Z',N')\<in> set ZNL\<rbrakk>\<Longrightarrow> N\<noteq>N'"
apply (simp add: set_conv_nth Set_of_Valid_ZNL_def);
apply clarify;
apply (drule_tac x=i in spec);
apply simp+;
apply clarify;
apply (drule CAN.neighbour_diff);
apply (drule sym, simp);
done;

(* each destination node is a CAN Node *)
lemma ZNL_Node_CAN_Node: "\<lbrakk>ZNL\<in> Set_of_Valid_ZNL C N Z;(a,b)\<in>set ZNL\<rbrakk>\<Longrightarrow> b\<in>CAN_Nodes C"
apply (simp add: Set_of_Valid_ZNL_def set_conv_nth);
apply clarify;
apply (rotate_tac -2, drule sym)
apply force;
done;

(* each destination node is inside its zone *)
lemma ZNL_NodeInZoneZi: "\<lbrakk>ZNL\<in> Set_of_Valid_ZNL C N Z;(a,b)\<in>set ZNL\<rbrakk>\<Longrightarrow> b\<in>NodesInZone C a"
apply (simp add: Set_of_Valid_ZNL_def set_conv_nth);
apply clarify;
apply (rotate_tac -2, drule sym)
apply force;
done;

(* each destination node is inside the source zone *)
lemma ZNL_NodeInZoneZ: "\<lbrakk>ZNL\<in> Set_of_Valid_ZNL C N Z;(a,b)\<in>set ZNL\<rbrakk>\<Longrightarrow> b\<in>NodesInZone C Z"
apply (frule_tac a=a in  ZNL_NodeInZoneZi);
apply auto;
apply (simp add: Set_of_Valid_ZNL_def set_conv_nth);
apply clarify;
apply (rotate_tac -2, drule sym)
apply (force simp: intersects_def);
done;

lemma ZNL_NodeInZoneZ_rev: "\<lbrakk>ZNL\<in> Set_of_Valid_ZNL C N Z;b\<in>NodesInZone C Z\<rbrakk>\<Longrightarrow>\<exists> Z' N.(Z',N)\<in>set ZNL\<and> (N=b\<or> b\<in> NodesInZone C (Z' - CAN_Zones C N))"
apply (simp add: Set_of_Valid_ZNL_def set_conv_nth intersects_def)
apply auto
apply (rule_tac x="fst (ZNL ! i)" in exI)
apply (rule_tac x="snd (ZNL ! i)" in exI)
apply auto
apply (rule inc_single_zone)
apply (auto simp: intersects_def)
done;

subsection{* a broadcast based on ZNL *}

types ZNLMapping ="Zone\<Rightarrow>nat\<Rightarrow>((Zone\<times>nat) list)"

text{*MARK*}
inductive ZNL_BC_msgs::"CAN\<Rightarrow>nat\<Rightarrow>nat\<Rightarrow>ZNLMapping\<Rightarrow>Message set\<Rightarrow>Message list\<Rightarrow>bool"
for C:: CAN and Mid::nat and init:: nat and znlmap::ZNLMapping   where 
BC_init: 
"ZNL_BC_msgs C Mid init znlmap {} [<Mid|init,init,Entire_Space-CAN_Zones C init>]"
| BC_step: 
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs (M#ML); M=<Mid' | s,d,Z>; 
 ML'=map (\<lambda>ZN. let Z'=fst ZN in let N'=snd ZN in <Mid'|d,N',Z'-CAN_Zones C N' >) 
     (znlmap Z d)\<rbrakk> \<Longrightarrow>
               ZNL_BC_msgs C Mid init znlmap (insert M msgs)  (ML@ML')"

abbreviation Messages_of_ZNL_BC::"CAN \<Rightarrow>nat \<Rightarrow>nat\<Rightarrow>ZNLMapping\<Rightarrow>Message set set"
where
"Messages_of_ZNL_BC C Mid init znlmap\<equiv> {msgset. ZNL_BC_msgs C Mid init znlmap msgset []}"

lemma finite_msgs_BC: "ZNL_BC_msgs C Mid init znlmap msgs ml \<Longrightarrow> finite msgs"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply auto
done
lemma  message_unchanged_pre[rule_format]: "ZNL_BC_msgs C Mid init znlmap msgs ml\<Longrightarrow> \<forall> m s d Z m' s' d' Z'.((<m|s,d,Z>\<in> msgs \<longrightarrow> <m'|s',d',Z'>\<in> msgs  \<longrightarrow> m=m') \<and>(<m|s,d,Z>\<in> set ml \<longrightarrow> <m'|s',d',Z'>\<in> msgs  \<longrightarrow> m=m')\<and> (<m|s,d,Z>\<in> set ml \<longrightarrow> <m'|s',d',Z'>\<in> set ml  \<longrightarrow> m=m'))"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply simp+;
 apply clarify;
 apply simp+
apply (intro impI conjI)
(*6*)
   apply (drule_tac x=m' in spec);
   apply (drule_tac x=s in spec);
   apply (drule_tac x=d in spec);
   apply (drule_tac x=Z in spec);
   apply simp+;
  apply force;
 apply (subgoal_tac "m=Mid'")
  apply simp;
  apply (drule_tac x=m in spec);
  apply (drule_tac x=s in spec);
  apply (drule_tac x=d in spec);
  apply (drule_tac x=Z in spec);
  apply simp+;
 apply force;
 apply (subgoal_tac "m'=Mid'");
  apply simp
(*4*)
  apply (drule_tac x=m in spec);
  apply (drule_tac x=sa in spec);
  apply (drule_tac x=da in spec);
  apply (drule_tac x=Za in spec);
  apply (drule_tac x=m' in spec);
  apply (drule_tac x=s in spec);
  apply (drule_tac x=d in spec);
  apply (drule_tac x=Z in spec);
  apply simp+;
 apply force;
apply (subgoal_tac "m=Mid'");
 apply simp;
 apply (drule_tac x=m in spec);
 apply (drule_tac x=s in spec);
 apply (drule_tac x=d in spec);
 apply (drule_tac x=Z in spec);
 apply simp+;
(*2*)
 apply force;
apply force;
done;


lemma  message_unchanged: "\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ml;<m|s,d,Z>\<in> msgs; <m'|s',d',Z'>\<in> msgs\<rbrakk>\<Longrightarrow>m=m'"
apply (insert message_unchanged_pre)
apply auto;
done;

lemma  message_unchanged_ML: "\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ml;<m|s,d,Z>\<in> set ml; <m'|s',d',Z'>\<in> set ml\<rbrakk>\<Longrightarrow>m=m'"
apply (insert message_unchanged_pre)
apply auto;
done;

definition HistoryZNL:: "CAN\<Rightarrow>nat\<Rightarrow>Message set\<Rightarrow>Message list\<Rightarrow>((Zone\<times>nat) set)"
where "HistoryZNL C init msgs ML= 
{(Entire_Space - (CAN_Zones C init),init)}
\<union>((\<lambda>msg. (Message_zone msg,Message_dest msg))`msgs)
\<union>(if ML=[] then{} else{(\<lambda>msg. (Message_zone msg,Message_dest msg))(ML!0)})"

lemma Historysimp: 
"ZN\<in>HistoryZNL C init msgs ML = 
((fst ZN=(Entire_Space - (CAN_Zones C init))  \<and> snd ZN=init)\<or>
(\<exists>M\<in>msgs. fst ZN=Message_zone M\<and>snd ZN=Message_dest M) \<or>
(ML\<noteq>[]\<and>(fst ZN=Message_zone (ML!0)\<and>snd ZN=Message_dest (ML!0))))"
apply (unfold HistoryZNL_def)
apply (case_tac ZN)
apply force
done


lemma ZNL_distinct_zones_1_Optimal[THEN mp]: " ZNL_BC_msgs C Mid init znlmap msgs ml\<Longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<longrightarrow> list_distinct_nodes C ml"
apply (erule broadcast.ZNL_BC_msgs.induct)
  apply rule
  apply simp
(*1*)
  apply auto
 apply (drule_tac x="(ac,ba)" in bspec)
  apply (force simp: Historysimp)
  apply force

(*4*)
 apply (drule_tac x="(ac,ba)" in bspec)
  apply (force simp: Historysimp)
  apply force
 apply (drule_tac x="(ac,ba)" in bspec)
  apply (force simp: Historysimp)
  apply force
(*2*)
apply (simp add: list_distinct_nodes_def2 Set_of_Optimal_ZNL_def);
 apply auto
 apply (drule_tac x="(Z,d)" in bspec)
  apply (force simp: Historysimp Message_zone_def Message_dest_def)

 apply clarify
 apply (rotate_tac -2,drule_tac x=i in spec,simp)
 apply clarify
 apply (rotate_tac -1,drule_tac x=j in spec,simp)
 apply (simp add: intersects_def Message_zone_def Let_unfold);
 apply auto
(*1*)
apply (simp add: list_message_nodes_nth intersects_def Message_zone_def Let_unfold)
apply (simp add: Set_of_Optimal_ZNL_def)
apply auto
apply (drule_tac x="(Z,a)" in bspec)
  apply (force simp: Historysimp Message_zone_def Message_dest_def)

apply clarify
apply (rotate_tac -2,drule_tac x=i in spec,simp)
apply force
done

(*destination of each message to be processed is not inside any zone of a message to be processed *) (* requires optimality *)
lemma ZNL_dest_outside_zone_Optimal[THEN mp]: 
"ZNL_BC_msgs C Mid init znlmap msgs ml\<Longrightarrow>(\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<longrightarrow>
 (\<forall> i<length ml.  Message_dest (ml!i)\<notin> list_message_nodes C ml)"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply rule
  apply (force simp: Message_dest_def  intersects_def list_message_nodes_nth Message_zone_def)
(*1*)
apply clarsimp
apply rule
 apply (case_tac  "i < length ML")
  apply (simp add: nth_append Message_dest_def)
(*3*)
  apply auto
 (*5*)
  apply (drule_tac x="(a,b)" in bspec)
  apply (force simp: Historysimp)
  apply force
  apply (drule_tac x="(a,b)" in bspec)
  apply (force simp: Historysimp)
  apply force

  apply (drule_tac x="Suc i" in spec,simp)
 apply (simp add: nth_append Message_dest_def)
 apply (frule ZNL_distinct_zones_1_Optimal)
 apply clarsimp
  apply (drule_tac x="(a,b)" in bspec)
  apply (force simp: Historysimp)
  apply force

 apply simp
 apply (simp add:  Set_of_Optimal_ZNL_def Let_unfold intersects_def)
 apply (drule_tac x="(Z,d)" in bspec)
  apply (force simp: Historysimp Message_zone_def Message_dest_def)

 apply clarify
 apply (rotate_tac -2)
 apply (drule_tac x="i-length ML" in spec)
 apply auto
 apply (simp add:  list_message_nodes_nth Message_zone_def );
 apply clarify
(*3*)
 apply (subgoal_tac "snd (znlmap Z d ! (i - length ML)) \<in> NodesInZone C (\<Union> i<length (znlmap Z d). fst (znlmap Z d ! i))")
  apply (subgoal_tac "snd (znlmap Z d ! (i - length ML)) \<in> {N \<in> CAN_Nodes C. \<exists>x. (\<exists>i<length ML. x \<in> (\<lambda>(m, s, d, Z). Z) (ML ! i)) \<and> x \<in> CAN_Zones C N}")
   apply (rotate_tac 3,drule_tac x="i-length ML" in spec)
   apply force
  apply force;
 apply (simp add: intersects_def);
 apply force
 apply (drule_tac x="(a,b)" in bspec)
  apply (force simp: Historysimp)
  apply force

(*1*)
apply (case_tac  "i < length ML");
 apply (simp add:  list_message_nodes_nth nth_append Message_zone_def Let_unfold)+;
 apply auto;
 apply (subgoal_tac "x\<in>Z")
  apply (drule ZNL_distinct_zones_1_Optimal)
  apply (simp add: list_message_nodes_nth Message_zone_def Message_dest_def)
  apply (case_tac "ML!i");
  apply auto
  apply (drule_tac x="(ab,ba)" in bspec)
  apply (force simp: Historysimp)
  apply force
(*3*)
   apply (simp add: list_message_nodes_nth Message_zone_def Message_dest_def)
   apply (drule_tac x="Suc i" in spec,clarsimp+)
   apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: intersects_def)
   apply (simp add: intersects_def)
(*2*)
   apply (drule_tac x="Suc i" in spec)
 apply (simp add:  Set_of_Optimal_ZNL_def Let_unfold intersects_def)
 apply (drule_tac x="(Z,a)" in bspec)
   apply (force simp: Historysimp Message_zone_def Message_dest_def)
 apply auto

(*1 else*)
apply (simp add:  list_message_nodes_nth nth_append Message_zone_def Let_unfold)+;
apply (simp add: Set_of_Optimal_ZNL_def);
apply auto
apply (drule_tac x="(Z,a)" in bspec)
   apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply clarify
apply (rotate_tac -2,frule_tac x=ia in spec); (* forall i<length znl*)
apply (drule_tac x="i - length ML" in spec,simp); (* forall i<length znl*)
apply clarify
apply (rotate_tac -2,drule_tac x="i-length ML" in spec); (* forall j<length znl*)
apply (case_tac "ia=  i - length ML")
 apply (auto simp: Message_dest_def)
apply (auto simp: intersects_def)
done

(*destination of each message already processed is not inside any zone of a message to be processed *) (* requires optimality *)
lemma ZNL_dest_outside_cover_zones_Optimal[THEN mp,THEN mp]: 
"ZNL_BC_msgs C Mid init znlmap msgs ml\<Longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) \<longrightarrow> M\<in> msgs \<longrightarrow> Message_dest M \<notin> list_message_nodes C ml"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply (force simp: Message_dest_def);
apply clarify;
apply (frule ZNL_dest_outside_zone_Optimal)
apply clarsimp
apply (drule_tac x="(a,b)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply force
   
   apply clarsimp
   apply (auto simp: Message_dest_def)
(*8*)
   apply (drule_tac x="(a,b)" in bspec)
   apply ( simp add: Historysimp )
   apply blast
   apply simp

   apply (drule_tac x=0 in spec,simp)
    apply auto
     apply (drule list_message_zone_CAN_Node)
     apply simp
    apply  (simp add: intersects_def Message_zone_def Set_of_Optimal_ZNL_def list_message_nodes_nth Let_unfold)
    apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
    apply force
   apply (drule list_message_zone_CAN_Node);
   apply simp
  apply  (simp add: Message_zone_def Set_of_Optimal_ZNL_def);
    apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
  apply auto
(*5*)
  apply (simp add: intersects_def Message_zone_def)
  apply (simp add: list_message_nodes_nth Let_unfold);
  apply clarify;
  apply (rotate_tac 3) (*\<forall>tup. tup \<in> CAN_Zones*)
  apply (drule_tac x=x in spec,simp)
  apply (force simp: Message_zone_def)
apply (drule_tac x="(a,b)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply blast
   apply force
apply (drule_tac x="(a,b)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply blast
   apply force
 apply (drule list_message_zone_CAN_Node)
 apply simp
(*1*)
apply  (simp add: Message_zone_def Set_of_Optimal_ZNL_def)
apply (drule_tac x="(Z,d)" in bspec)
   apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply auto
apply (simp add: intersects_def Message_zone_def)
apply (simp add: list_message_nodes_nth Let_unfold);
apply clarify
apply (rotate_tac 4) (*\<forall>tup. tup \<in> CAN_Zones*)
apply (force simp: Message_zone_def)
done

(* \<lbrakk>\<forall>N Z. znlmap1 Z N \<in> Set_of_Optimal_ZNL C1 N Z; ZNL_BC_msgs C1 Mid1 init1 znlmap1 msgs1 ml1; M1 \<in> msgs1\<rbrakk> *)
(* \<Longrightarrow> Message_dest M1 \<notin> list_message_nodes C1 ml1 *)

(* requires optimality -- in practice disjoint is sufficient *)
lemma  unique_message_pre[THEN mp]: 
"ZNL_BC_msgs C Mid init znlmap msgs ml\<Longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) 
\<longrightarrow>(\<forall> m s d Z m'  Z'.(
  (<m|s,d,Z>\<in> msgs \<longrightarrow> <m'|s,d,Z'>\<in> msgs  \<longrightarrow> (m=m' \<and> Z=Z')) \<and>
  (<m|s,d,Z>\<in> set ml \<longrightarrow> <m'|s,d,Z'>\<notin> msgs)\<and> 
  (<m|s,d,Z>\<in> set ml \<longrightarrow> <m'|s,d,Z'>\<in> set ml  \<longrightarrow> (m=m' \<and>Z=Z'))))
\<and>distinct ml"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply clarsimp

apply (subgoal_tac "ZNL_BC_msgs C Mid init znlmap (insert M msgs) (ML@ML')")
defer;
 apply (erule_tac Mid'=Mid' and s=s and d=d and Z=Z and znlmap=znlmap in BC_step)
   apply assumption+
apply (intro allI impI conjI)
(*6*)
      apply clarify
      apply (rotate_tac 2, erule_tac s=sa and d=da and s'=sa and d'=da and Z=Za in message_unchanged)
       apply fastsimp
      apply (fastsimp)
     apply simp
     apply (elim disjE)
(*8*)
      apply force
     apply simp+
     apply (erule impE)
    apply (clarsimp)
    apply (drule_tac x="(ac,ba)" in bspec)
    apply (force simp: Historysimp)
    apply simp+
(*6*)     
     apply (erule impE)
    apply (clarsimp)
    apply (drule_tac x="(ac,ba)" in bspec)
    apply (force simp: Historysimp)
    apply simp+
(*5*)    
     apply (erule impE)
    apply (clarsimp)
    apply (drule_tac x="(ac,ba)" in bspec)
    apply (force simp: Historysimp)
    apply simp+
(*4*)
  apply (elim disjE)
   apply clarsimp
    apply (erule impE)
    apply (clarsimp)
    apply (drule_tac x="(a,b)" in bspec)
    apply (force simp: Historysimp)
    apply clarsimp+
(*5*)
   apply (drule_tac x=Mid' in spec);
   apply (rotate_tac -1,drule_tac x=s in spec);
   apply (rotate_tac -1,drule_tac x=d in spec);
   apply (rotate_tac -1,drule_tac x=Z in spec);
   apply (rotate_tac -1,drule_tac x=m in spec);
   apply (rotate_tac -1,drule_tac x=Za in spec)
   apply simp
  apply clarify
  apply rule

(*5*)
  apply clarify;
  apply simp
  apply clarify
  apply (drule_tac x="(Z,ba)" in bspec)
    apply (simp only: Historysimp Message_dest_def Message_zone_def)
    apply force
    

  apply (drule Optimal_isValid_ZNL,drule ZNL_diff)
   apply force
  apply force
 apply simp
 apply clarify;
 apply simp;
 apply (frule_tac x="(Z,d)" in bspec)
   apply (simp(no_asm) add: Historysimp Message_dest_def Message_zone_def)

 apply (drule Optimal_isValid_ZNL)
 apply (drule ZNL_NodeInZoneZ)
(*5*)
   apply simp
  apply (drule ZNL_dest_outside_cover_zones_Optimal)
    apply (clarsimp)
    apply (drule_tac x="(a,b)" in bspec)
    apply (simp only: Historysimp)
    apply (erule disjE,force)
    apply (erule disjE,force)
    apply force
apply simp+
(*4*)    
  apply (simp add: Message_dest_def Message_zone_def)
 apply (rotate_tac 2, erule_tac s=sa and d=da and s'=sa and d'=da and Z=Za in message_unchanged_ML) 
                   (* on   ZNL_BC_msgs C Mid init znlmap (insert M msgs) (ML @ ML'); < m| sa, da, Za> \<in> set (ML @ ML'); < m'| sa, da, Z'> \<in> set (ML @ ML')*)
  apply simp+
(*2*)
 apply (elim disjE)+
    apply clarsimp
    apply (erule impE)
    apply (clarsimp)
    apply (drule_tac x="(a,b)" in bspec)
    apply (force simp: Historysimp)
    apply clarsimp+
(*4*)
    apply ( drule ZNL_dest_outside_zone_Optimal)
    apply (clarsimp)
    apply (thin_tac ?A)
    apply (drule_tac x="(a,b)" in bspec)
    apply (force simp: Historysimp)
    apply simp
    apply (clarsimp simp: list_distinct_nodes_def2 set_conv_nth Message_dest_def)
    apply (drule_tac x="Suc i" in spec)
    apply simp+
    apply (case_tac "ML!i",clarsimp simp: MessagesAndPaths.list_message_nodes_nth)
    apply (subgoal_tac "c\<in>CAN_Nodes C")
    apply simp+
(*5*)
    apply (simp add: Message_zone_def)
    apply (subgoal_tac "znlmap Z b\<in>Set_of_Optimal_ZNL C b Z")
    apply (thin_tac ?A)
    apply (thin_tac ?A)
    apply (thin_tac ?A)
    apply (simp add: Set_of_Optimal_ZNL_def set_conv_nth Let_unfold)
    apply (clarify,drule_tac x=ia in spec)
    apply (drule sym)
    apply (force simp: intersects_def)
    apply (thin_tac ?A)
    apply (drule_tac x="(Z,b)" in bspec)
    apply (force simp: Historysimp Message_zone_def Message_dest_def)
    apply force

    apply (thin_tac ?A)
    apply (drule_tac x="(Z,b)" in bspec)
    apply (force simp: Historysimp Message_zone_def Message_dest_def)
    apply (simp add: Set_of_Optimal_ZNL_def set_conv_nth Let_unfold)
    apply (clarify,drule_tac x=ia in spec)
    apply (drule sym, force)

(*3*)
  apply (drule ZNL_dest_outside_zone_Optimal)
  apply clarsimp
  apply (thin_tac ?A)
  apply (drule_tac x="(ad,bb)" in bspec)
    apply (force simp: Historysimp )
  apply simp+
  apply (simp add: set_conv_nth);
  apply clarify;
  apply (rotate_tac -8,drule_tac x="Suc i" in spec,simp);
  apply clarify
  apply (rotate_tac 4,drule sym)
  apply (simp add: Message_dest_def Message_zone_def)
  apply (rotate_tac 2)
    apply (thin_tac ?A)

 apply (simp add: Set_of_Optimal_ZNL_def set_conv_nth Let_unfold);
 apply clarify
 apply (drule_tac x="(Z,d)" in bspec) 
     apply (force simp: Historysimp Message_zone_def Message_dest_def)
 apply clarsimp+
 apply (drule_tac x=ia in spec)
 apply clarsimp
 apply (drule sym)
 apply clarsimp
 apply (force simp: intersects_def)
(*2*)
    apply (thin_tac ?A)
    apply (thin_tac ?A)

 apply (simp add: Set_of_Optimal_ZNL_def set_conv_nth Let_unfold);
 apply clarify
 apply (drule_tac x="(Z,d)" in bspec) 
     apply (force simp: Historysimp Message_zone_def Message_dest_def)
 apply clarsimp+
 apply (frule_tac x=ia in spec)
 apply (drule_tac x=i in spec)
 apply clarsimp
 apply force

(*1*)
apply auto
 apply (drule_tac x="(ac,ba)" in bspec) 
     apply (force simp: Historysimp )
     apply force
 apply (drule_tac x="(ac,ba)" in bspec) 
     apply (force simp: Historysimp )
     apply force
 apply (drule_tac x="(ac,ba)" in bspec) 
     apply (force simp: Historysimp )
     apply force
(*2*)
 apply (drule_tac x="(Z,d)" in bspec) 
     apply (force simp: Historysimp Message_zone_def Message_dest_def)
   apply (frule Optimal_isValid_ZNL)
     apply (drule distinct_ZNL_Zones_Optimal)
     apply (drule distinct_ZNL_Nodes)
     apply (auto simp: distinct_conv_nth Let_unfold)

apply (frule ZNL_dest_outside_zone_Optimal)
  apply clarsimp
  apply (thin_tac ?A)
  apply (thin_tac ?A)
  apply (drule_tac x="(aa,ba)" in bspec)
  apply (thin_tac ?A)
    apply ( simp add: Historysimp )
    apply blast
    apply simp
(*1*)
apply (simp add: set_conv_nth)
apply clarify
apply (subgoal_tac "b\<in>NodesInZone C Z")
 apply ( simp add: Message_dest_def Message_zone_def intersects_def)
 apply (rotate_tac -5, drule sym, drule sym)
 apply (rotate_tac -3, drule_tac x="Suc i" in spec,simp+)
  apply (drule_tac x="(Z,d)" in bspec) 
     apply (force simp: Historysimp Message_zone_def Message_dest_def)
 apply clarsimp+
(*1*)
  apply (drule_tac x="(Z,d)" in bspec) 
     apply (force simp: Historysimp Message_zone_def Message_dest_def)
 apply clarsimp+
 apply (drule Optimal_isValid_ZNL)
apply (drule broadcast.ZNL_NodeInZoneZ)
apply (auto simp: set_conv_nth)
done

(*requires optimality -- disjoint zones is sufficient *)
lemma  unique_message_Z: "\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs  ml;<m|s,d,Z>\<in> msgs; <m'|s,d,Z'>\<in> msgs; \<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)\<rbrakk>
\<Longrightarrow>Z=Z'"
apply (insert unique_message_pre)
apply auto
done


(*if a node is in the zone of a treated message, it is either destination of a treated message or of a message to be treated or in the zones to be covered by a message to be treated *)


(*if a node is in the zone of a treated message, it is either destination of a treated message or of a message to be treated or in the zones to be covered by a message to be treated *)

lemma message_treated_coverage[rule_format]:
"ZNL_BC_msgs C Mid init znlmap msgs  ml\<Longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Valid_ZNL C (snd ZN) (fst ZN)) \<longrightarrow>
(\<forall>  s d Z m'  Z'.
  (<Mid|s,d,Z>\<in> msgs \<longrightarrow>
  (N \<in> NodesInZone C Z) \<longrightarrow>
       ((\<exists>  s' Z'.<Mid|s',N,Z'>\<in> msgs) \<or>
        (\<exists>  s' Z'.<Mid|s',N,Z'>\<in> set ml) \<or>
        (\<exists>  s' d' Z'.(<Mid|s',d',Z'>\<in> set ml \<and> N \<in> (NodesInZone C Z')))
       )))
"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply simp+
apply clarify
apply rule
 apply (intro conjI impI)
 apply (rule disjI2)
 apply (drule_tac x="(Z,d)" in bspec)
  apply (force simp: Historysimp Message_zone_def Message_dest_def)
  apply simp+
 apply (drule_tac b=N in ZNL_NodeInZoneZ_rev)
  apply force
 apply (elim exE conjE disjE)
(*3*)
  apply (rule disjI1)
  apply clarify
  apply simp+
  apply (rule_tac x=d in exI)
  apply (rule_tac x="Z' -  CAN_Zones C N" in exI)
  apply force
(*2*)
 apply (rule disjI2)
 apply clarify
 apply simp+
 apply (rule_tac x=d in exI)
 apply (rule_tac x=Na in exI)
 apply (rule_tac x="Z' -  CAN_Zones C Na" in exI)
 apply force
(*1*)
apply (intro conjI impI)
apply (erule impE)
  apply clarsimp
    apply (drule_tac x="(a,b)" in bspec)
    apply (force simp: Historysimp)
    apply simp
apply (rotate_tac 1,drule_tac x=sa in spec)
apply (rotate_tac -1,drule_tac x=da in spec)
apply (rotate_tac -1,drule_tac x=Za in spec)
apply (erule impE,force)+
apply (elim disjE)
  apply force
(*2*)
 apply (elim exE conjE disjE)
  apply simp+
  apply (rule disjI1)
  apply force
 apply (rule disjI2,rule disjI1)
 apply force
apply (elim exE conjE disjE)
defer
 apply (rule disjI2,rule disjI2)
 apply force

(* copy paste from the begining \<rightarrow> improve? *)
 apply simp+
 apply (rule disjI2)
 apply (drule_tac x="(Z,d)" in bspec)
  apply (force simp: Historysimp Message_zone_def Message_dest_def)
  apply simp+
 apply (drule_tac b=N in ZNL_NodeInZoneZ_rev)
  apply force
 apply (elim exE conjE disjE)
(*2*)
  apply (rule disjI1)
  apply clarify
  apply simp+
  apply (rule_tac x=d in exI)
  apply (rule_tac x="Z'a -  CAN_Zones C N" in exI)
  apply force
(*1*)
 apply (rule disjI2)
 apply clarify
 apply simp+
 apply (rule_tac x=d in exI)
 apply (rule_tac x=Na in exI)
 apply (rule_tac x="Z'a -  CAN_Zones C Na" in exI)
 apply force
done

lemma insert_iff_same: "\<lbrakk>a\<in>A;a\<notin>B;a\<noteq>b\<rbrakk>\<Longrightarrow>insert b A \<noteq> insert b B "
apply force done

lemma ZNL_of_treated_messages[rule_format]: 
"ZNL_BC_msgs C Mid init znlmap msgs ml \<Longrightarrow> \<forall>m s d Z. <m|s,d,Z>\<in>msgs \<longrightarrow> 
 set (map (\<lambda>ZN. let Z'=fst ZN in let N'=snd ZN in <m | d,N',Z'- CAN_Zones C N' >) (znlmap Z d)) \<subseteq> msgs\<union> (set ml)"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply simp
apply (intro allI impI)
apply simp
apply (erule disjE)
apply rule
apply simp+
apply (drule_tac x=m in spec)
apply (drule_tac x=sa in spec)
apply (drule_tac x=da in spec)
apply (drule_tac x=Za in spec)
apply force
done

lemma ZNL_of_initial_message[rule_format]: 
"ZNL_BC_msgs C Mid init znlmap msgs ml \<Longrightarrow> 
 (\<lambda>ZN. let Z'=fst ZN in let N'=snd ZN in <Mid | init,N',Z'- CAN_Zones C N' >) (Entire_Space, init) \<in> msgs\<union> (set ml)"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply auto
done

lemma final_treated_messages[rule_format]:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ml ; ZNL_BC_msgs C Mid init znlmap Finalmsgs []\<rbrakk>\<Longrightarrow> (msgs\<union>set ml) \<subseteq>  Finalmsgs"
apply (insert ZNL_of_initial_message[of C Mid init znlmap Finalmsgs])
apply (drule_tac x="[]" in meta_spec)
apply simp
apply (insert ZNL_of_treated_messages[of C Mid init znlmap Finalmsgs])
apply (drule_tac x="[]" in meta_spec)
apply simp
apply (erule broadcast.ZNL_BC_msgs.induct)
apply auto
apply (drule_tac x=Mid' in meta_spec)
apply (drule_tac x=s in meta_spec)
apply (drule_tac x=d in meta_spec)
apply (drule_tac x=Z in meta_spec)
apply auto
done


lemma coverage:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap Finalmsgs []; N\<in>CAN_Nodes C; init\<in>CAN_Nodes C;
   (\<forall> ZN\<in>(HistoryZNL C init Finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Valid_ZNL C (snd ZN) (fst ZN))\<rbrakk>\<Longrightarrow>
       \<exists> s Z. <Mid|s, N, Z>\<in> Finalmsgs"
apply (case_tac "N=init")
apply (frule ZNL_of_initial_message)
apply force
apply (frule Entire_space_nodes)
apply (frule_tac N'=init in NodesInZone_substract)
apply simp+
apply (frule ZNL_of_initial_message)
apply simp+
apply (case_tac "N=init")
apply simp+
apply (drule_tac s=init and d=init and N=N and Z="Entire_Space - CAN_Zones C init"in message_treated_coverage)
apply clarsimp+
done

lemma ZNL_msgs_neighbour[rule_format]:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ML;init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>(\<forall> ZN\<in>(HistoryZNL C init msgs ML). znlmap (fst ZN) (snd ZN) \<in> Set_of_Valid_ZNL C (snd ZN) (fst ZN))\<longrightarrow>(\<forall>x y m Z. < m| x, y, Z> \<in> msgs\<union> set ML\<longrightarrow>(y=init\<or>(CAN_neighbour C x y)))"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply (unfold Set_of_Valid_ZNL_def)
apply clarsimp
apply clarify
(*1*)
apply (erule impE)
apply clarsimp
apply (frule_tac x="(a,b)" in bspec)
apply (thin_tac ?A)
apply (thin_tac ?A)
apply (thin_tac ?A)
apply (force simp: Historysimp)
apply force
apply (drule_tac x="(Z,d)" in bspec)
apply (force simp: Historysimp Message_dest_def Message_zone_def)
apply auto
apply (simp add: set_conv_nth, clarify)
apply (rotate_tac -3,drule_tac x=i in spec)
apply (rotate_tac 1, drule sym)
apply force
done

lemma ZNL_msgs_Nodes[rule_format]:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ML;init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>(\<forall> ZN\<in>(HistoryZNL C init msgs ML). znlmap (fst ZN) (snd ZN) \<in> Set_of_Valid_ZNL C (snd ZN) (fst ZN))\<longrightarrow>(\<forall>x y m Z. < m| x, y, Z> \<in> msgs\<union> set ML\<longrightarrow>(x\<in>CAN_Nodes C\<and>y\<in>CAN_Nodes C))"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply simp
(*1*)
apply clarify+
apply (erule impE)
apply clarsimp
apply (frule_tac x="(a,b)" in bspec)
apply (thin_tac ?A)
apply (thin_tac ?A)
apply (thin_tac ?A)
apply (force simp: Historysimp)
apply force
(*1*)
apply (subgoal_tac "d\<in>CAN_Nodes C")
apply ( drule_tac x=x in spec)
apply (rotate_tac -1, drule_tac x=y in spec)
apply (rotate_tac -1, drule_tac x=m in spec)
apply (rotate_tac -1, drule_tac x=Za in spec)
apply clarsimp
apply (erule disjE)
apply simp+
(*2*)
apply (drule_tac x="(Z,d)" in bspec)
apply ( simp add: Historysimp Message_dest_def Message_zone_def)
apply (unfold Set_of_Valid_ZNL_def)
apply clarsimp
apply (clarsimp simp: set_conv_nth)
apply (rotate_tac -4,drule_tac x=i in spec)
apply (simp+)
apply (rotate_tac 1, drule sym)
apply force
(*1*)
apply ( drule_tac x=s in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=Mid' in spec)
apply (rotate_tac -1, drule_tac x=Z in spec)
apply force
done

lemma ZNL_msgs_path[rule_format]:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ML;init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>(\<forall> ZN\<in>(HistoryZNL C init msgs ML). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<longrightarrow>(\<forall>x y m Z. < m| x, y, Z> \<in> msgs\<union> set ML\<longrightarrow>((\<exists> MsgL. valid_path msgs MsgL\<and>destination MsgL=x\<and>source MsgL=init) \<or> x=init))" 
apply (erule broadcast.ZNL_BC_msgs.induct)
apply force
apply (intro allI impI)
apply simp+
apply (erule impE)
apply clarsimp
apply (frule_tac x="(ac,ba)" in bspec)
apply (simp add: Historysimp)
apply blast
apply force
(*1*)
apply (elim disjE)
apply (rotate_tac 2, drule_tac x=s in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=Mid' in spec)
apply (rotate_tac -1, drule_tac x=Z in spec)
apply clarsimp
apply (rule_tac x=MsgL in exI)
apply (simp add: valid_path_def)
apply clarsimp
apply (rotate_tac 2, drule_tac x=x in spec)
apply (rotate_tac -1, drule_tac x=y in spec)
apply (rotate_tac -1, drule_tac x=m in spec)
apply (rotate_tac -1, drule_tac x=Za in spec)
apply clarsimp
apply (rule_tac x=MsgL in exI)
apply (simp add: valid_path_def)
(*2*)
apply (rotate_tac 2, drule_tac x=x in spec)
apply (rotate_tac -1, drule_tac x=y in spec)
apply (rotate_tac -1, drule_tac x=m in spec)
apply (rotate_tac -1, drule_tac x=Za in spec)
apply clarsimp
apply (rule_tac x=MsgL in exI)
apply (simp add: valid_path_def)
apply clarsimp
apply (rotate_tac 2, drule_tac x=s in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=Mid' in spec)
apply (rotate_tac -1, drule_tac x=Z in spec)
apply clarsimp
apply (elim disjE)
apply clarsimp
apply (rule_tac x="MsgL@[< Mid' | Message_dest (MsgL ! (length MsgL - Suc 0)), d, Z>]" in exI)
apply (frule unique_message_pre)
(*3*)
apply clarsimp
apply (frule_tac x="(a,b)" in bspec)
apply (force simp: Historysimp)
apply force
(*2*)
apply (drule_tac x="(Z,d)" in bspec)
apply ( simp(no_asm) add: Historysimp Message_dest_def Message_zone_def)
(*2*)
apply (simp add: set_conv_nth valid_path_def nth_append)
apply (simp add: Set_of_Optimal_ZNL_def)
apply clarsimp
apply (rotate_tac 5, drule_tac x=i in spec)
apply clarsimp
apply rule+
apply (subgoal_tac "ia = length MsgL -  1")
apply (simp add: Message_dest_def Message_source_def)
apply force
(*2*)
apply rule+
apply clarsimp
apply (drule_tac x=ia in spec)
apply simp
apply (rotate_tac 6, drule_tac x=Mid' in spec)
apply (rotate_tac -1, drule_tac x="Message_dest (MsgL ! (length MsgL - Suc 0))" in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=Z in spec)
apply simp+
apply (rotate_tac -1, drule_tac x=Mid' in spec)
apply (rotate_tac -1, drule_tac x=Z in spec)
apply clarsimp
apply (drule_tac x=0 in spec)+
apply force
(*2*)
apply (simp add: Message_dest_def Message_source_def)
apply clarsimp            
apply (rule_tac x="[< Mid' | init, d, Z>]" in exI)
apply auto
apply (simp add: Message_dest_def Message_source_def)+
done




theorem ZNL_BC:
"\<lbrakk> ZNL_BC_msgs C Mid init znlmap Finalmsgs []; init\<in>CAN_Nodes C;
 (\<forall> ZN\<in>(HistoryZNL C init Finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<rbrakk>\<Longrightarrow>
     (C, Finalmsgs, init)\<in>Broadcast"
apply (unfold Broadcast_def)
apply (frule unique_message_pre)
 apply force
apply (subgoal_tac "(\<forall> ZN\<in>(HistoryZNL C init Finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Valid_ZNL C (snd ZN) (fst ZN))")
 apply auto
(*7*)
   apply force
  apply force
 apply (frule ZNL_msgs_Nodes)
    apply auto
(*4*)
   apply (frule ZNL_msgs_Nodes)
      apply auto
  apply (frule ZNL_msgs_neighbour)
     apply auto
 apply (drule ZNL_msgs_path)
(*5*)
 apply auto
apply (rule Optimal_isValid_ZNL)
apply force
done



theorem coverage_ZNL:
"\<lbrakk> ZNL_BC_msgs C Mid init znlmap Finalmsgs []; init\<in>CAN_Nodes C;
  (\<forall> ZN\<in>(HistoryZNL C init Finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<rbrakk>\<Longrightarrow>
     Coverage (<C, Finalmsgs, init>)"
apply (subgoal_tac "Rep_Broadcast (<C, Finalmsgs, init>)\<in>Broadcast")
defer apply (rule Rep_Broadcast)
apply (subgoal_tac "Rep_Broadcast (<C, Finalmsgs, init>) = (C,Finalmsgs, init)")
apply (unfold Coverage_def)
apply auto
apply (auto simp: BC_CAN_def BC_msgs_def BC_initiator_def)
apply (case_tac "n=init")
apply simp
apply (drule_tac  N=n in coverage)
apply auto
(*2*)
apply (rule Optimal_isValid_ZNL)
apply auto
apply (drule ZNL_BC)
apply auto
apply (rotate_tac -1)
apply (drule Abs_Broadcast_inverse)
apply auto
done

section {*towards existence of a ZNL*}

lemma ZNL_existence_pre_0: 
"\<lbrakk>N \<in>CAN_Nodes C;Connected C (Z\<union> CAN_Zones C N \<union> Z');NodesInZone C Z' = {N'}; N'\<in>CAN_Nodes C;N'\<notin>NodesInZone C Z;N'\<noteq>N\<rbrakk> 
\<Longrightarrow>  (CAN_neighbour C N N') \<or> (\<exists> N''\<in>NodesInZone C Z. CAN_neighbour C N' N'')"
(is "\<lbrakk>?nnode;?connectedZZp;?npinZp;?npnode;?npnotinZ;?nnotnp\<rbrakk>\<Longrightarrow>?neighbournnp\<or>?existsNp")
proof -
  assume ?nnode ?connectedZZp ?npinZp ?npnode ?npnotinZ ?nnotnp
  from `?nnode` have "(CAN_Zones C N )\<noteq>{}"
    by (rule zone_non_empty)
  then obtain x where "x\<in>(CAN_Zones C N )" by auto
  with `?nnode` have "N\<in>NodesInZone C (CAN_Zones C N )" 
    by  (rule_tac x=x in NodesInZoneI, auto) 
  with NodesInZoneInc[of "(CAN_Zones C N )" "(Z\<union>CAN_Zones C N \<union> Z')"] have "N\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')"  by auto 
  from `?npinZp` NodesInZoneInc[of Z' "(Z\<union>CAN_Zones C N \<union> Z')"] have "N'\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')"  by auto
  with `?nnotnp` and `N\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')`  finite_intersects[of C"(Z\<union>CAN_Zones C N \<union> Z')"]  
    have "1 < (ZoneSize C (Z\<union> CAN_Zones C N \<union> Z'))" by (intro  finite_non_singleton, auto)
  with `?npnode` Connected_exists_neighbour[of C "(Z \<union> CAN_Zones C N \<union> Z')"] `?connectedZZp` `?npinZp`
    have  "\<exists>N''\<in>CAN_Nodes C. N''\<in>NodesInZone C (Z\<union> CAN_Zones C N \<union> Z') \<and> CAN_neighbour C N' N''" 
      proof -
        from `?connectedZZp` and  `1 < (ZoneSize C (Z\<union> CAN_Zones C N \<union> Z'))` 
              Connected_exists_neighbour[of C "(Z \<union> CAN_Zones C N \<union> Z')"] 
            have "\<forall>Na\<in>NodesInZone C (Z \<union> CAN_Zones C N \<union> Z'). \<exists>N'\<in>CAN_Nodes C. N' \<in> NodesInZone C (Z \<union> CAN_Zones C N \<union> Z') \<and> CAN_neighbour C Na N' " 
              by auto
      moreover from `N'\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')` `?npinZp` have "N'\<in> NodesInZone C (Z \<union> CAN_Zones C N \<union> Z')" by auto
      ultimately show ?thesis by auto 
    qed
    then obtain  N'' where "N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')" "CAN_neighbour C N' N''" by auto
    thus ?thesis
      proof cases
        assume "N''=N"
        with `N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')` `CAN_neighbour C N' N''` show  ?thesis by (metis neighbour_refl)
      next
        assume "N''\<noteq>N"
        from neighbour_diff[of C N' N''] and `N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')` `CAN_neighbour C N' N''` have "N'\<noteq>N''"by auto
        with `N''\<noteq>N` `?npinZp` `N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')` `CAN_neighbour C N' N''` have "N''\<in>NodesInZone C Z"
        proof -
          from `N'\<noteq>N''` have "N''\<noteq>N'" ..
          from `?npinZp`have "Z'\<subseteq>CAN_Zones C N'" by (rule NodesInZoneSingleton)
          hence  "(Z\<union>CAN_Zones C N \<union> Z') \<subseteq> (Z\<union>CAN_Zones C N \<union> CAN_Zones C N')" by blast
          with `N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union> Z')` have "N''\<in>NodesInZone C (Z\<union>CAN_Zones C N \<union>CAN_Zones C N')" by (auto simp: intersects_def)
          from this `?npnode``N''\<noteq>N'` have "N''\<in>NodesInZone C ((Z\<union>CAN_Zones C N\<union>CAN_Zones C N')-CAN_Zones C N')" by (rule NodesInZone_substract)
          hence "N''\<in>NodesInZone C (Z\<union>CAN_Zones C N)" by (auto simp: intersects_def)
          from this `?nnode``N''\<noteq>N` have "N''\<in>NodesInZone C ((Z\<union>CAN_Zones C N)-CAN_Zones C N)" by (rule NodesInZone_substract)
          thus "N''\<in>NodesInZone C Z" by (auto simp: intersects_def)
        qed
        with `CAN_neighbour C N' N''` show ?thesis
          by auto
      qed
    qed;







lemma Valid_NoTwoZonesConnected_ImpliesOptimal[rule_format]:
"\<forall> Z.  ZNL\<in>Set_of_Valid_ZNL C N Z \<longrightarrow>( \<forall> i j. ((i<length ZNL\<and> j<length ZNL\<and>i\<noteq>j)\<longrightarrow> NoNodeInBothZones C (fst (ZNL!i)) (fst (ZNL!j))) )\<longrightarrow> ZNL\<in>Set_of_Optimal_ZNL C N Z"
apply (induct_tac ZNL)
apply auto
apply (simp add: Set_of_Optimal_ZNL_def)+
apply (drule_tac x="Z-a" in spec)
apply (erule impE)
apply (simp add: Set_of_Valid_ZNL_def)+
apply rule
apply rule
apply rule
apply clarify
apply (rotate_tac 1,drule_tac x="Suc i" in spec)
apply force
apply (subgoal_tac "(\<Union> i<length list. fst (list ! i))\<inter>a={}" )
apply (subgoal_tac "Z=(\<Union> i<length list. fst (list ! i))\<union>a" )
apply simp+
apply force
(*3*)
apply (subgoal_tac " Z=(\<Union> i<Suc (length list). fst (((a, b) # list) ! i))")
apply simp+
apply (subgoal_tac "(length list)\<le> (length (map fst list))")
apply (drule_tac a=a in Union_unfold)
apply (subgoal_tac "(length list)\<le> (length (list))")
apply (drule_tac F=fst in map_Union)
apply (subgoal_tac "Suc (length list)\<le> (length  ((a, b) # list))")
apply (drule_tac F=fst in map_Union)
apply simp+
(*2*)
apply (subgoal_tac "\<forall>i<length list.  fst (list ! i) \<inter> a={}")
apply force
apply clarify
apply (drule_tac x=0 in spec)
apply (rotate_tac -1, drule_tac x="Suc i" in spec)
apply simp+
apply (drule NoNodeInBothZones_disjoint)
apply force

apply (simp add: Set_of_Valid_ZNL_def)+
done




lemma Zone_decomposition: "\<exists> ZL. ((\<Union> i<length ZL. ZL!i) = Z) \<and> (\<forall>i<length ZL. (Connected C (ZL!i))) \<and> ((\<forall> i<length ZL. \<forall> j<length ZL. (i\<noteq>j\<longrightarrow>\<not>Connected C ((ZL!i)\<union>(ZL!j)))))"
apply (rule_tac P="\<lambda>Z. \<exists> ZL. ((\<Union> i<length ZL. ZL!i) = Z) \<and> (\<forall>i<length ZL. (Connected C (ZL!i))) \<and> ((\<forall> i<length ZL. \<forall> j<length ZL. (i\<noteq>j\<longrightarrow>\<not>Connected C ((ZL!i)\<union>(ZL!j)))))" in induct_node_zone_2)
apply auto
apply (rule_tac x="[]" in exI)
apply auto
apply (rule_tac x="(Z'\<union>\<Union>{Z\<in>set ZL. Connected C (Z\<union> Z')})#filter (\<lambda> Z. \<not>Connected C (Z\<union> Z')) ZL" in exI)
apply auto
apply (case_tac xa)
apply (simp add: set_conv_nth)+
apply clarify
apply simp+
apply (rule_tac x=i in bexI)
apply force

apply force
(*4*)
apply simp+
apply (drule filter_nth)
apply clarify
apply (rule_tac x=i in bexI)
apply force
apply force

apply (case_tac "Connected C ((ZL!xa)\<union> Z')")

apply (rule_tac x=0 in bexI)
apply force
apply force
apply (rotate_tac -1)
apply (drule_tac P="\<lambda>Z.\<not> Connected C (Z \<union> Z')" in filter_nth_rev)
apply simp+
apply clarify
apply (rule_tac x="Suc i" in bexI)
apply simp+
apply (case_tac i)
apply simp+
apply (subgoal_tac "(\<Union>{Z \<in> set ZL. Connected C (Z \<union> Z')})= (\<Union>i<length (filter (\<lambda> Z. Connected C (Z\<union>Z')) ZL).(filter (\<lambda> Z. Connected C (Z\<union>Z')) ZL)!i) ")
apply simp
apply (rule Connected_Union)
apply (subgoal_tac "Connected C (([Z\<leftarrow>ZL . Connected C (Z \<union> Z')] ! ia)\<union> Z')")
(*8*)
apply (subst Un_commute)
apply (force simp: set_conv_nth)
apply (subgoal_tac "[Z\<leftarrow>ZL . Connected C (Z \<union> Z')] ! ia\<in> set [Z\<leftarrow>ZL . Connected C (Z \<union> Z')]")
apply force
apply (simp add: set_conv_nth)
apply blast
apply (force simp: intersects_def)
apply (drule filter_nth)
apply force
apply (rule_tac N=N in SingleNodeConnected)
apply simp+

apply rule
apply rule
apply simp+
apply clarify

apply (simp add: set_conv_nth)
apply clarify
apply (rotate_tac -3,drule_tac j=i and ZL=ZL in filter_nth_rev)
apply simp+
apply force

apply clarify
apply simp+
apply (rotate_tac -1, frule filter_nth)
apply clarify
apply (rule_tac x="[Z\<leftarrow>ZL . Connected C (Z \<union> Z')] ! ia" in exI)
apply (simp add: set_conv_nth)
apply rule
apply force
(*3*)
apply (rotate_tac -1,drule sym)
apply (subgoal_tac "[Z\<leftarrow>ZL . Connected C (Z \<union> Z')] ! ia\<in> set [Z\<leftarrow>ZL . Connected C (Z \<union> Z')]")
   (* the method to prove filter P ... !ia verifies P! *)
apply force
apply ( simp add: set_conv_nth)
apply blast

apply simp+
apply (drule filter_nth)
apply force

apply (case_tac i)
apply simp+
apply (case_tac j)
apply simp+
apply (frule filter_nth)
apply clarify
(*2*)
apply (drule Connected_union_Union_2)
apply (erule SingleNodeConnected)
apply simp+
apply (force simp: set_conv_nth)
apply simp

apply safe

apply (subgoal_tac "\<not> Connected C (Z' \<union> [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat)")
apply force
apply (subgoal_tac "\<not> Connected C ( [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat \<union> Z')")
apply (subst Un_commute)
apply simp
apply (subgoal_tac "  [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat\<in> set [Z\<leftarrow>ZL . \<not>Connected C (Z \<union> Z')]")
apply force
apply ( simp add: set_conv_nth)
apply blast
(*2*)
apply ( simp add: set_conv_nth)

apply clarify
apply (rotate_tac -2, frule filter_nth_rev) (* ZL!i \<union> Z'*)
apply force
apply clarify
apply (case_tac "i=ia")
apply (subgoal_tac "[Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat \<in> set [Z\<leftarrow>ZL . \<not>Connected C (Z \<union> Z')]")
apply force
apply (simp add: set_conv_nth)
apply (rule_tac x=nat in exI)
apply force
apply force

(*2*)
apply simp+
apply (case_tac j)
(* similar to above *)
apply simp+
apply (frule filter_nth)
apply clarify
apply (subgoal_tac "([Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat \<union> (Z' \<union> \<Union>{Z \<in> set ZL. Connected C (Z \<union> Z')})) = ( (Z' \<union> \<Union>{Z \<in> set ZL. Connected C (Z \<union> Z')}) \<union>[Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat)")
apply simp
apply (drule Connected_union_Union_2)
apply (erule SingleNodeConnected)
apply simp+
apply (force simp: set_conv_nth)
apply simp

apply safe
(*3*)
apply (subgoal_tac "\<not> Connected C (Z' \<union> [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat)")
apply force
apply (subgoal_tac "\<not> Connected C ( [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat \<union> Z')")
apply (subst Un_commute)
apply simp
apply (subgoal_tac "  [Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat\<in> set [Z\<leftarrow>ZL . \<not>Connected C (Z \<union> Z')]")
apply force
apply ( simp add: set_conv_nth)
apply blast

apply ( simp add: set_conv_nth)

apply clarify
apply (rotate_tac -2, frule filter_nth_rev) (* ZL!i \<union> Z'*)
apply force
apply clarify
apply (case_tac "i=ia")
apply (subgoal_tac "[Z\<leftarrow>ZL . \<not> Connected C (Z \<union> Z')] ! nat \<in> set [Z\<leftarrow>ZL . \<not>Connected C (Z \<union> Z')]")
apply force
apply (simp add: set_conv_nth)
apply (rule_tac x=nat in exI)
apply force
apply force

apply simp+

apply (rotate_tac 2,erule contrapos_pp)

apply (drule_tac k=nata in filter_nth_double)
apply simp+
apply force
done


lemma OptimalZNL_existence:
"\<lbrakk>Connected C (Z\<union>(CAN_Zones C N));N\<in>CAN_Nodes C ; N\<notin>NodesInZone C Z\<rbrakk>\<Longrightarrow>\<exists> ZNL. ZNL\<in>Set_of_Optimal_ZNL C N Z"
apply (insert Zone_decomposition[of Z C])
apply (case_tac "Z={}")
apply simp
apply (rule_tac x="[]" in exI)
apply (simp add: Set_of_Optimal_ZNL_def)

apply (elim exE conjE)
apply (subgoal_tac "\<forall> i<length ZL. \<exists> N'\<in>NodesInZone C (ZL!i). CAN_neighbour C N N'")
apply (subgoal_tac "\<exists> NL. (length NL = length ZL \<and> (\<forall> i<length ZL. NL!i\<in>NodesInZone C (ZL!i)\<and> CAN_neighbour C N (NL!i)))")
apply (elim exE conjE)
apply (rule_tac x="zip ZL NL" in exI)
apply auto
apply (simp add: Set_of_Optimal_ZNL_def)
apply auto
(*4*)
apply (subgoal_tac "\<not> Connected C (ZL!ia \<union> ZL!j)")
apply (drule not_connected_union_no_intersection)
apply simp+
apply (rotate_tac -1,drule_tac x="NL!ia" in spec)
apply simp+
apply force
apply force

apply (subgoal_tac "\<not> Connected C (ZL!ia \<union> ZL!j)")
apply (drule not_connected_union_no_intersection)
apply simp+

apply (rotate_tac -3)
apply (erule contrapos_pp)
apply simp
apply (rule_tac P="\<forall>i<length ZL. \<exists>x. x \<in> CAN_Nodes C \<and> intersects (CAN_Zones C x) (ZL ! i) \<and> CAN_neighbour C N x" in mp)
apply (induct_tac ZL)
apply simp+
apply clarify
apply (erule impE)
apply clarify
apply (rotate_tac -2, drule_tac x="Suc ia" in spec)
apply force
apply (rotate_tac -2, drule_tac x=0 in spec)
apply clarify
apply simp
apply clarify
(*3*)
apply(rule_tac x="xa#NL" in exI)
apply simp+
apply clarify
apply (case_tac ia)
apply force
apply force
apply assumption
(*1*)
apply (subgoal_tac "ZL!i\<noteq>{}")
apply auto
apply (rotate_tac -1)
apply (frule_tac C=C in CAN.intersects_empty_rev)
apply clarsimp
apply (simp add: intersect_node_simp) 
apply (subgoal_tac "(\<Union> i<length ZL. ZL ! i) \<union> CAN_Zones C N=CAN_Zones C N\<union>(\<Union> i<length ZL. ZL ! i) ")
apply simp
apply (drule_tac i=i and N'=node in connected_Union_neighbour)
apply simp+
apply (force simp: intersects_def)
apply simp+
apply clarsimp
apply blast
done

lemma BC_Optimal_disjoint_Zones[rule_format]:
" \<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs MList; init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>
  (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) \<longrightarrow>
(\<forall>i<length MList. \<forall> j<length MList. i\<noteq>j \<longrightarrow> NoNodeInBothZones C (Message_zone (MList!i))   (Message_zone (MList!j))) 
"apply (erule broadcast.ZNL_BC_msgs.induct)
apply clarsimp

(*1*)
apply clarsimp
apply (erule impE)
apply clarsimp
apply (drule_tac x="(a,b)" in bspec)
apply (force simp: Historysimp)
apply simp

(*1*)
apply (unfold Set_of_Optimal_ZNL_def Message_zone_def)
apply (case_tac "i< length ML")
apply (case_tac "j< length ML")

apply simp+

(*3*)
apply (drule_tac x="Suc i" in spec)
apply simp+
apply (rotate_tac -1, drule_tac x="Suc j" in spec)
apply simp
apply (force simp: nth_append)

apply (simp add: nth_append)

apply (rotate_tac -4)
apply (drule_tac x="Suc i" in spec)
apply simp+
apply (rotate_tac -1, drule_tac x=0 in spec)
apply simp

apply (drule_tac x="(Z,d)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
(*2*)
apply clarsimp

apply (case_tac "(znlmap Z d ! (j - length ML))")

apply auto
apply (subgoal_tac "a-CAN_Zones C b\<subseteq>  Z ")

apply (drule_tac Z''="(CAN_Zones C x)" in intersect_included)
apply force
apply (subgoal_tac "x\<in> NodesInZone C ((\<lambda>(m, s, d, Z). Z) (ML ! i)) \<inter> NodesInZone C Z")
apply blast
apply blast
(*2*)
apply (subgoal_tac "a\<subseteq>Z")
apply blast

apply (subgoal_tac "a=fst (znlmap Z d ! (j - length ML))")
apply force
apply simp
(*1*)
apply (case_tac "j< length ML")
(* same thing inverting i and j *)

apply (auto simp: nth_append)
apply (rotate_tac -7)
apply (drule_tac x="0" in spec)
apply simp+
apply (rotate_tac -1, drule_tac x="Suc j" in spec)
apply simp

apply (drule_tac x="(Z,d)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply clarsimp

apply (case_tac "(znlmap Z d ! (i - length ML))")

apply auto
apply (subgoal_tac "a-CAN_Zones C b\<subseteq>  Z ")

apply (drule_tac Z''="(CAN_Zones C x)" in intersect_included)
apply force
apply (subgoal_tac "x\<in> NodesInZone C ((\<lambda>(m, s, d, Z). Z) (ML ! i)) \<inter> NodesInZone C Z")
apply blast
apply blast

apply (subgoal_tac "a\<subseteq>Z")
apply blast
(*2*)
apply (subgoal_tac "a=fst (znlmap Z d ! (i - length ML))")
apply force
apply simp

apply (drule_tac x="(Z,d)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply clarsimp

apply (case_tac "(znlmap Z d ! (i - length ML))")
apply (case_tac "(znlmap Z d ! (j - length ML))")
apply clarsimp
(*1*)
apply (rotate_tac -5, drule_tac x="(i - length ML)" in spec)
apply (erule impE)
apply arith
apply clarsimp
apply (rotate_tac -1,drule_tac x="(j - length ML)" in spec)
apply (erule impE)
apply arith
apply (erule impE)
apply arith
apply simp

apply (subgoal_tac "x\<in> NodesInZone C a \<inter> NodesInZone C aa")
apply blast
apply rule

apply (subgoal_tac "(a - CAN_Zones C b)\<subseteq>a")
apply (simp)
apply (erule intersect_included)
apply simp+
apply blast
(*1*)
apply (subgoal_tac "(aa - CAN_Zones C ba)\<subseteq>aa")
apply (simp)
apply (erule intersect_included)
apply simp+
apply blast
done

(*crucial theorem: reused several times for proving most of the theorems below: broadcast can do one step, i.e. keep the same messages, but add one message to the treated messages, also shows that the zone to be treated gets smaller \<Rightarrow> it will converge thanks to the ordering defined below *)
theorem progress_ZNL[rule_format]:
"\<lbrakk> ZNL_BC_msgs C Mid init znlmap msgs MList; init\<in>CAN_Nodes C;
    (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN));MList\<noteq>[] \<rbrakk>\<Longrightarrow>
     \<exists> MList'. msgs \<union> set MList \<subseteq> msgs \<union>{MList!0}\<union> set MList' \<and>
(MList'=[]\<or>(length MList>1\<and>MList'!0= MList!1)\<or>((znlmap (Message_zone (MList!0))(Message_dest (MList!0)))\<noteq>[]\<and>(let ZN = (znlmap (Message_zone (MList!0))(Message_dest (MList!0)))!0 in (Message_dest (MList'!0)=snd ZN\<and>Message_zone(MList'!0)=(fst ZN)- CAN_Zones C (snd ZN)))))\<and>
ZNL_BC_msgs C Mid init znlmap (msgs\<union>{MList!0}) MList'\<and>((\<Union>{NodesInZone C Zi|Zi. (Zi\<in>set (map Message_zone MList'))}\<subset>\<Union>{NodesInZone C Zi|Zi. (Zi\<in>set (map Message_zone MList))})\<or>(\<exists> a. MList=a#MList'\<and>Message_zone a ={}))"

apply (case_tac MList)
apply (simp)

apply (insert BC_Optimal_disjoint_Zones[of C Mid init znlmap msgs MList])

apply clarsimp
apply (frule_tac Mid'=a and s=aa and d=ab and Z= b in  BC_step,simp+)
apply (rule_tac x="(list @ map (\<lambda>ZN. let N' = snd ZN in < a| ab, N', fst ZN - CAN_Zones C N'>) (znlmap b ab))" in exI)
apply simp+
apply (subgoal_tac "(znlmap b ab=[]) = (b={}) ")
apply rule
apply force
apply rule
apply force
apply rule
apply simp+
apply (clarsimp simp: nth_append Let_unfold Message_zone_def Message_dest_def)
apply (auto simp: Message_zone_def)
(*5*)
apply (drule_tac x="NodesInZone C ba" in spec)
apply simp
apply blast

apply (drule_tac x="NodesInZone C b" in spec)
apply (subgoal_tac "(af-CAN_Zones C bb)\<subseteq>b")
apply (rotate_tac -1,drule_tac Z''="(CAN_Zones C x)" in intersect_included)
apply simp+
apply blast
apply (clarsimp simp: Set_of_Optimal_ZNL_def set_conv_nth)
apply (rotate_tac -2, drule sym)
apply (drule_tac x="(b,ab)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply force

(*3*)
apply (drule_tac x="(b,ab)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply (clarsimp simp: Set_of_Optimal_ZNL_def set_conv_nth)
apply (case_tac "znlmap b ab")
apply blast
apply clarsimp
apply (subgoal_tac "ba\<in>\<Union>{NodesInZone C Zi | Zi. Zi = (\<Union> i<Suc (length lista). fst (((ac, ba) # lista) ! i)) \<or> (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")
apply (subgoal_tac " ba\<notin>\<Union>{(NodesInZone C Zi)|Zi.
          (Zi \<in> (\<lambda>(m, s, d, Z). Z) `
                (set list  \<union>
                 (\<lambda>ZN. let N' = snd ZN in < a| ab, N', fst ZN - CAN_Zones C N'>) ` set ((ac, ba) # lista) ))}")
apply (simp add: set_conv_nth)
apply (thin_tac "\<Union>?A=\<Union>?B")
prefer 2
apply (thin_tac "\<Union>?A=\<Union>?B")
prefer 2
apply clarsimp

(*4*)
apply (rotate_tac -3, erule disjE)
apply (force simp: intersects_def)

apply clarsimp
apply (rotate_tac -1, erule disjE)
apply (clarsimp simp: set_conv_nth)

(*5*)
apply (drule_tac x=0 in meta_spec)
apply (drule_tac x="Suc ib" in meta_spec)
apply simp+
apply (subgoal_tac "ba\<in>NodesInZone C (\<Union> i<Suc (length lista). fst (((ac, ba) # lista) ! i))")
apply (subgoal_tac "ba\<in>NodesInZone C ((\<lambda>(m, s, d, Z). Z) (list ! ib))")
apply blast
apply (drule_tac t="list!ib" in sym)
apply (force)
apply simp
(*5*)
apply (subgoal_tac "Suc (length lista)\<le> length ((ac, ba) # lista)")
apply (drule_tac Z="(CAN_Zones C ba)" and f="fst" in intersects_Union_list)
apply simp
apply (rule_tac x=0 in exI)
apply (drule_tac x=0 in spec)
apply force
apply simp
(*4*)
apply (clarsimp simp: set_conv_nth)

apply (drule_tac x=0 in spec)
apply clarsimp
apply (drule_tac x="Suc ib" in spec)
apply simp+
apply (drule_tac t="lista!ib" in sym)
apply clarsimp
apply (subgoal_tac "ba\<in>NodesInZone C ac")
apply (subgoal_tac "ba\<in>NodesInZone C ag")
apply blast
(*5*)
apply (subgoal_tac "ag-CAN_Zones C bb\<subseteq> ag")
apply simp
apply (rotate_tac -1, erule intersect_included)
apply force
apply blast
apply blast

(*3*)
apply (subgoal_tac "ba\<in>NodesInZone C ac")

apply clarsimp
apply (rule_tac x="NodesInZone C (\<Union> i<Suc (length lista). fst (((ac, ba) # lista) ! i))" in exI)
apply (clarsimp)
apply rule
apply blast
apply (subgoal_tac "Suc (length lista)\<le> length ((ac, ba) # lista)")
(*5*)
apply (drule_tac Z="(CAN_Zones C ba)" and f="fst" in intersects_Union_list)
apply simp
apply (rule_tac x=0 in exI)
apply (drule_tac x=0 in spec)
apply force
apply simp+
apply (drule_tac x=0 in spec)
apply force

(*2*)
apply (drule_tac x="(b,ab)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply (force simp: Set_of_Optimal_ZNL_def)

apply (drule_tac x="({},ab)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
(*1*)
apply (clarsimp simp: Set_of_Optimal_ZNL_def)
apply (force simp: intersects_def)
done

lemma wf_BC_order: "wf (inv_image ( less_than<*lex*>less_than)  (\<lambda> L. (ZoneSize C (\<Union> (set (map Message_zone L))),length L)))"
apply auto
done

(*lemma wf_BC_order: "wf (inv_image (({(MList1, MList2). ZoneSize C (\<Union> (set (map Message_zone MList1)))<ZoneSize C (\<Union> (set (map Message_zone MList2)))}<*lex*>{(MList1,MList2). (length MList1<length MList2)}) ) (\<lambda> L. (L,L)))"
apply auto
apply (subgoal_tac "{(MList1, MList2). ZoneSize C (\<Union>x\<in>set MList1. Message_zone x) < ZoneSize C (\<Union>x\<in>set MList2. Message_zone x)} = (inv_image less_than (\<lambda> L. ZoneSize C (\<Union> (set (map Message_zone L)))))")
apply simp
apply simp
apply (simp add: inv_image_def)
apply (subgoal_tac "{(MList1, MList2). (length MList1 < length MList2)}= (inv_image less_than length )")
apply (rotate_tac -1, erule ssubst)
apply simp+
apply (simp add: inv_image_def)
done*)


(* FINAL THEOREM: NOT USED IN NEXT STEPS *)
theorem finish_ZNL_No_History[rule_format]:
"\<forall> msgs.  ZNL_BC_msgs C Mid init znlmap msgs MList\<longrightarrow> init\<in>CAN_Nodes C\<longrightarrow>
  (\<forall> N Z. znlmap Z N \<in> Set_of_Optimal_ZNL C N Z)\<longrightarrow>
     (\<exists> finalmsgs. ZNL_BC_msgs C Mid init znlmap finalmsgs [])"
apply (insert wf_BC_order[of C])
apply (erule_tac P="\<lambda> MList.\<forall>msgs. ZNL_BC_msgs C Mid init znlmap msgs MList \<longrightarrow>
           init \<in> CAN_Nodes C \<longrightarrow> (\<forall>N Z. znlmap Z N \<in> Set_of_Optimal_ZNL C N Z) \<longrightarrow> (\<exists>finalmsgs. ZNL_BC_msgs C Mid init znlmap finalmsgs [])" in wf_induct)
(* \<lbrakk>wf r; \<And>x. \<forall>y. (y, x) \<in> r \<longrightarrow> P y \<Longrightarrow> P x\<rbrakk> \<Longrightarrow> P a *)

apply clarsimp
apply (case_tac x)
apply force
apply simp+
apply (frule progress_ZNL)
apply (simp,simp,simp)
(*1*)
apply clarify
apply (thin_tac   "MList' = [] \<or>?B \<or>?C")
apply (elim disjE conjE)
apply (drule_tac x=MList' in spec)
apply (elim conjE)
apply (erule impE)
(*3*)
apply (subgoal_tac 
" {N \<in> CAN_Nodes C. intersects (CAN_Zones C N) (Message_zone (< a| aa, ab, b>)) \<or> intersects (CAN_Zones C N) (\<Union>x\<in>set list. Message_zone x)}= NodesInZone C (\<Union>{Zi. Zi = Message_zone (< a| aa, ab, b>) \<or> Zi \<in> Message_zone ` set list})")
apply (rotate_tac -1, erule ssubst)
apply (rule ZoneSize_included)
apply simp+
(*4*)
apply (subgoal_tac "NodesInZone C (\<Union>x\<in>set MList'. Message_zone x) = \<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> Message_zone ` set MList')}")
apply simp
apply (subgoal_tac "{N \<in> CAN_Nodes C. \<exists>x. (x = Message_zone (< a| aa, ab, b>) \<or> x \<in> Message_zone ` set list) \<and> intersects (CAN_Zones C N) x} = \<Union>{(NodesInZone C Zi)|Zi. (Zi = Message_zone (< a| aa, ab, b>) \<or> Zi \<in> Message_zone ` set list)}")
apply simp
apply (unfold Message_zone_def)
apply simp+
apply rule
apply rule
(*6*)
apply clarsimp
apply (rotate_tac -4, erule disjE)
apply simp
apply (drule_tac x= "NodesInZone C b" in spec)
apply simp+
apply (drule_tac x= b in spec)
apply blast
apply (clarsimp simp: set_conv_nth)
apply (drule_tac x= "NodesInZone C ba" in spec)
apply simp+
apply (drule_tac x= ba in spec)
apply blast
(*5*)
apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')} \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = b \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")
apply clarsimp
apply (rule_tac x= Zi in exI)
apply simp

apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')}
        \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = (\<lambda>(m, s, d, Z). Z) (< a| aa, ab, b>) \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")
apply (simp add: set_conv_nth )
apply (subgoal_tac "(\<Union> x\<in>{(MList' ! i)| i. (i < length MList')}. (\<lambda>(m, s, d, Z). Z) x) = \<Union>{(\<lambda>(m, s, d, Z). Z) (MList'!i)|i. i<length MList'}")
apply (simp only: NodesInZone_Union)
apply clarsimp
apply blast
(*4*)
apply rule
apply clarsimp
apply (rule_tac x=ba in exI)
apply simp
apply (rule_tac x=i in exI)
apply force
apply rule
apply clarsimp
apply (rule_tac x=xb in exI)
apply (rule_tac x=ac in exI)
apply (rule_tac x=ad in exI)
apply (rule_tac x=ba in exI)
apply ( thin_tac "msgs \<subseteq>  ?B")
apply (rotate_tac 5)
apply (thin_tac ?A)
apply force
(*3*)
apply (thin_tac "msgs \<union> ?A\<subseteq> msgs \<union> ?B \<union>set MList'")
apply simp+
apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')}
        \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = b \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")

apply rule
apply clarsimp
apply (erule disjE)
apply blast
apply (subgoal_tac "(\<Union>x\<in>set list. ((\<lambda>(m, s, d, Z). Z) x)) = \<Union>{(\<lambda>(m, s, d, Z). Z) x|x. x\<in> set list}")
apply simp
apply clarsimp
apply (rule_tac x=xa in exI)
apply simp+
apply blast
(*4*)
apply blast

apply clarsimp
apply (subgoal_tac "(\<Union>x\<in>set list. ((\<lambda>(m, s, d, Z). Z) x)) = \<Union>{(\<lambda>(m, s, d, Z). Z) x|x. x\<in> set list}")
apply force
apply force

apply force

(*1*)
apply (drule_tac x=list in spec)
apply clarsimp
apply (rotate_tac -1, erule impE)
apply (force simp: intersects_def)
apply force
done

lemma NodesInZone_simp[simp]:
"(N\<in>NodesInZone C Z) = (N\<in>CAN_Nodes C \<and>intersects (CAN_Zones C N) Z)"
apply (simp add: intersects_def)
done

lemma Connected_union_singleNode:
"\<lbrakk>intersects (CAN_Zones C N) Z;N\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>Connected C (Z\<union>(CAN_Zones C N)) = Connected C Z"
apply rule
apply (clarsimp simp: Connected_def)
apply (drule_tac x=n in spec)
apply (erule impE)
apply simp+
(*2*)
apply (drule_tac x=n' in spec)
apply clarsimp
apply (rule_tac x=node_list in exI)
apply (auto simp: intersects_def)
apply (drule_tac x=i in spec)
apply clarsimp+
apply (drule_tac N'="(node_list ! i)" in CAN.inc_single_zone_2)
apply (auto simp: intersects_def)
(*1*)
apply (clarsimp simp: Connected_def)
apply (drule_tac x=n in spec)
apply (erule impE)
apply (erule disjE)
apply simp+
apply (drule_tac N'=n in CAN.inc_single_zone_2)
apply simp+
apply (force simp: intersects_def)
apply (force simp: intersects_def)
(*1*)
apply (drule_tac x=n' in spec)
apply clarsimp
apply (erule impE)
apply (erule_tac P="intersects (CAN_Zones C n') Z " in disjE)
apply simp+
apply (drule_tac N'=n' in CAN.inc_single_zone_2)
apply simp+
apply (force simp: intersects_def)
apply (force simp: intersects_def)
apply auto
done

axioms Connected_Space: "Connected C Entire_Space"
-- "This is an axiom: proving it requires reasoning on geometrical arguments; it requires additional requirements on the CAN definition (linking neighbours and geometry"

lemma BC_Optimal_Connected_Zones[rule_format]:
" \<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs MList; init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow>
  (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) \<longrightarrow>
  (\<forall> M \<in>set MList. Connected C (Message_zone M \<union> CAN_Zones C (Message_dest M)))"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply clarsimp
apply (simp add: Set_of_Optimal_ZNL_def Message_dest_def Message_zone_def set_conv_nth)
apply (subgoal_tac "intersects (CAN_Zones C init) Entire_Space" )
apply (insert Connected_Space[of C])
apply (drule_tac Connected_union_singleNode)
apply simp+
apply (drule CAN.Entire_space_nodes)
apply simp
apply simp+
(*1*)
apply (clarsimp simp: Message_zone_def Message_dest_def)
apply (erule disjE)
apply (erule impE)
apply clarsimp
apply (drule_tac x="(ac,ba)" in bspec)
apply (force simp: Historysimp)
apply simp
(*2*)
apply force

apply (drule_tac x="(Z,d)" in bspec)
apply (simp add: Historysimp Message_zone_def Message_dest_def)
(*1*)
apply (simp add: Set_of_Optimal_ZNL_def Message_dest_def Message_zone_def set_conv_nth Let_unfold)
apply clarsimp
apply (drule_tac x=i in spec)
apply (case_tac "(znlmap Z d ! i)")
apply clarsimp
apply (drule_tac Connected_union_singleNode)
apply simp+
done

lemma BC_ZNL_only_sensible_to_history_pre[rule_format]:
"ZNL_BC_msgs C Mid init znlmap msgs MList\<Longrightarrow>
  ZNL_BC_msgs C Mid init znlmap msgs MList\<longrightarrow>
    (\<forall> ZN\<in>(HistoryZNL C init msgs []). znlmap' (fst ZN) (snd ZN) = znlmap (fst ZN) (snd ZN))\<longrightarrow>  ZNL_BC_msgs C Mid init znlmap' msgs MList"
apply (erule broadcast.ZNL_BC_msgs.induct)
apply auto
apply (insert BC_init)
apply simp+
apply (drule_tac x="(a,b)" in bspec)
apply (force simp: Historysimp)
apply simp
apply (rule BC_step)
apply auto
apply (drule_tac x="(Z,d)" in bspec)
apply (force simp: Historysimp Message_zone_def Message_dest_def)
apply simp
done

lemma BC_ZNL_only_sensible_to_history:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs MList;
    (\<forall> ZN\<in>(HistoryZNL C init msgs []). znlmap' (fst ZN) (snd ZN) = znlmap (fst ZN) (snd ZN))\<rbrakk>\<Longrightarrow>
  ZNL_BC_msgs C Mid init znlmap' msgs MList"
apply (rule BC_ZNL_only_sensible_to_history_pre)
apply auto
done


(*Now find a new znlmap so that the new ZN added still ensure optimal history (we could add here condition on znlmap wrt znlmap' but is it crucial?) 
USES Optimal_ZNL_existence*)
theorem progress_ZNL_2[rule_format]:
"\<lbrakk> ZNL_BC_msgs C Mid init znlmap msgs MList; init\<in>CAN_Nodes C;
    (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN));MList\<noteq>[] \<rbrakk>\<Longrightarrow>
     \<exists>  MList' znlmap'. msgs \<union> set MList \<subseteq> msgs \<union>{MList!0}\<union> set MList' \<and>
(ZNL_BC_msgs C Mid init znlmap' (msgs\<union>{MList!0}) MList'\<and>
 ((\<Union>{NodesInZone C Zi|Zi. (Zi\<in>set (map Message_zone MList'))}\<subset>\<Union>{NodesInZone C Zi|Zi. (Zi\<in>set (map Message_zone MList))})
   \<or>(\<exists> a. MList=a#MList'\<and>Message_zone a ={}))
 \<and>(\<forall> ZN\<in>(HistoryZNL C init (msgs\<union>{MList!0}) MList'). znlmap' (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)))"
apply (frule progress_ZNL) 
apply simp+
apply clarsimp
apply (rule_tac x=MList' in exI)
apply simp+
apply (case_tac MList')
apply (rule_tac x=znlmap in exI)
apply (clarsimp simp: Historysimp)
apply (drule_tac x="(a,b)" in bspec)
apply (simp add: Historysimp)
apply force
apply simp
apply (subgoal_tac "Message_dest a\<in>CAN_Nodes C")
apply (subgoal_tac "Connected C ( (Message_zone a)\<union>CAN_Zones C (Message_dest (a)))")
apply (drule_tac OptimalZNL_existence)
apply simp+
apply (drule ZNL_dest_outside_zone_Optimal)
apply (clarsimp simp: Historysimp)
apply (erule disjE)
apply (drule_tac x="1" in spec)
apply (simp add: Message_dest_def Message_zone_def)
apply (clarsimp simp: list_message_nodes_nth intersects_def)
apply (case_tac MList, clarsimp+)
apply (drule_tac x=tup in spec)
apply (force simp: Message_zone_def)
apply (clarsimp simp: Message_dest_def Message_zone_def)
apply (case_tac MList)
apply simp
apply clarsimp
apply (case_tac "znlmap ba ae")
apply simp
apply clarsimp
apply (force simp: intersects_def)

apply (clarsimp simp: Message_dest_def Message_zone_def)
(*3*)
apply (case_tac "(b,ab)\<in>HistoryZNL C init ({MList ! 0} \<union> msgs) []")
(* this should not happen in fact *)
apply (rule_tac x=znlmap in exI)
apply simp+
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply (drule_tac x="(ac,ba)" in bspec)
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply simp

(*now the case that really happens*)
apply (rule_tac x="\<lambda> Z N. if Z=b \<and> N=ab then ZNL else znlmap Z N" in exI)
apply simp+
apply rule
apply (erule BC_ZNL_only_sensible_to_history)
apply clarsimp
(*3*)
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply (case_tac MList)
apply simp
apply clarsimp

apply (rotate_tac -5,erule disjE)
apply (drule_tac x="(ac,ba)" in bspec)
apply (simp add: Historysimp)
apply simp
(*3*)
apply (rotate_tac -1,erule disjE)
apply (drule_tac x="(ac,ba)" in bspec)
apply (force simp: Historysimp Message_dest_def Message_zone_def)
apply simp
(*3*)
apply (rotate_tac -1,erule disjE)
apply (drule_tac x="(ac,ba)" in bspec)
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply clarify
apply force
apply simp
(*2*)
apply simp
apply (erule disjE)
apply (erule broadcast.BC_Optimal_Connected_Zones)
apply simp+
apply (case_tac MList)
apply simp
apply (clarsimp simp: Message_zone_def Message_dest_def)
apply (case_tac "znlmap ba ae")
apply simp
apply clarsimp
apply (drule_tac x="(ba,ae)" in bspec)
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply (simp add: Set_of_Optimal_ZNL_def)
apply clarsimp
apply (rotate_tac -1, drule_tac x=0 in spec)
apply clarsimp
apply (drule Connected_union_singleNode)
apply simp+
apply (case_tac a)
apply (clarsimp simp: Message_dest_def)
apply (erule disjE)
apply (drule_tac m=aaa and x=ba and y=c and Z=d in broadcast.ZNL_msgs_Nodes)
apply simp
apply clarsimp
apply (drule_tac x="(a,b)" in bspec)
apply simp+
apply (erule Optimal_isValid_ZNL)
apply simp+
(*1*)
apply (case_tac MList)
apply simp
apply (clarsimp simp: Message_zone_def Message_dest_def)
apply (drule_tac x="(b,ab)" in bspec)
apply (clarsimp simp: Historysimp Message_dest_def Message_zone_def)
apply (simp add: Set_of_Optimal_ZNL_def)
apply clarsimp
apply (rotate_tac -1, drule_tac x=0 in spec)
apply (case_tac "znlmap b ab")
apply auto
done


(* THEN by recurrence we can use progress to reach a final state ensuring optimality*)
theorem finish_ZNL_existence[rule_format]:
"\<forall> msgs znlmap.  ZNL_BC_msgs C Mid init znlmap msgs MList\<longrightarrow> init\<in>CAN_Nodes C\<longrightarrow>
  (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN))\<longrightarrow>
     (\<exists> finalmsgs znlmap'. ZNL_BC_msgs C Mid init znlmap' finalmsgs [] \<and>
  (\<forall> ZN\<in>(HistoryZNL C init finalmsgs []). znlmap' (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)))"
apply (insert wf_BC_order[of C])
apply (erule_tac P="\<lambda> MList.\<forall>msgs znlmap. ZNL_BC_msgs C Mid init znlmap msgs MList \<longrightarrow>
           init \<in> CAN_Nodes C \<longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs MList). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) \<longrightarrow> (\<exists>finalmsgs znlmap'. ZNL_BC_msgs C Mid init znlmap' finalmsgs []\<and>
  (\<forall> ZN\<in>(HistoryZNL C init finalmsgs []). znlmap' (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)))" in wf_induct)
(* \<lbrakk>wf r; \<And>x. \<forall>y. (y, x) \<in> r \<longrightarrow> P y \<Longrightarrow> P x\<rbrakk> \<Longrightarrow> P a *)

apply clarsimp
apply (case_tac x)
apply force
apply simp+
apply (frule progress_ZNL_2)
apply (simp,simp,simp)
(*1*)
apply clarify
apply (elim disjE conjE)
apply (drule_tac x=MList' in spec)
apply (elim conjE)
apply (erule impE)
(*3*)
apply (subgoal_tac 
" {N \<in> CAN_Nodes C. intersects (CAN_Zones C N) (Message_zone (< a| aa, ab, b>)) \<or> intersects (CAN_Zones C N) (\<Union>x\<in>set list. Message_zone x)}= NodesInZone C (\<Union>{Zi. Zi = Message_zone (< a| aa, ab, b>) \<or> Zi \<in> Message_zone ` set list})")
apply (rotate_tac -1, erule ssubst)
apply (rule ZoneSize_included)
apply simp+
(*4*)
apply (subgoal_tac "NodesInZone C (\<Union>x\<in>set MList'. Message_zone x) = \<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> Message_zone ` set MList')}")
apply simp
apply (subgoal_tac "{N \<in> CAN_Nodes C. \<exists>x. (x = Message_zone (< a| aa, ab, b>) \<or> x \<in> Message_zone ` set list) \<and> intersects (CAN_Zones C N) x} = \<Union>{(NodesInZone C Zi)|Zi. (Zi = Message_zone (< a| aa, ab, b>) \<or> Zi \<in> Message_zone ` set list)}")
apply simp
apply (unfold Message_zone_def)
apply simp+
apply rule
apply rule
(*6*)
apply clarsimp
apply (rotate_tac -4, erule disjE)
apply simp
apply (drule_tac x= "NodesInZone C b" in spec)
apply simp+
apply (drule_tac x= b in spec)
apply blast
apply (clarsimp simp: set_conv_nth)
apply (drule_tac x= "NodesInZone C ba" in spec)
apply simp+
apply (drule_tac x= ba in spec)
apply blast
(*5*)
apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')} \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = b \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")
apply clarsimp
apply (rule_tac x= Zi in exI)
apply simp

apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')}
        \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = (\<lambda>(m, s, d, Z). Z) (< a| aa, ab, b>) \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")
apply (simp add: set_conv_nth )
apply (subgoal_tac "(\<Union> x\<in>{(MList' ! i)| i. (i < length MList')}. (\<lambda>(m, s, d, Z). Z) x) = \<Union>{(\<lambda>(m, s, d, Z). Z) (MList'!i)|i. i<length MList'}")
apply (simp only: NodesInZone_Union)
apply clarsimp
apply (rotate_tac 5, thin_tac ?A)
apply blast
(*4*)
apply rule
apply clarsimp
apply (rule_tac x=ba in exI)
apply simp
apply (rule_tac x=i in exI)
apply force
apply rule
apply clarsimp
apply (rule_tac x=xb in exI)
apply (rule_tac x=ac in exI)
apply (rule_tac x=ad in exI)
apply (rule_tac x=ba in exI)
apply ( thin_tac "msgs \<subseteq>  ?B")
apply (rotate_tac 6)
apply (thin_tac ?A)
apply simp+
apply (rule_tac x=i in exI)
apply force
(*3*)
apply (thin_tac "msgs \<union> ?A\<subseteq> msgs \<union> ?B \<union>set MList'")
apply simp+
apply (thin_tac "\<Union>{(NodesInZone C Zi)|Zi. (Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set MList')}
        \<subset> \<Union>{(NodesInZone C Zi)|Zi. (Zi = b \<or> Zi \<in> (\<lambda>(m, s, d, Z). Z) ` set list)}")

apply rule
apply clarsimp
apply (erule disjE)
apply blast
apply (subgoal_tac "(\<Union>x\<in>set list. ((\<lambda>(m, s, d, Z). Z) x)) = \<Union>{(\<lambda>(m, s, d, Z). Z) x|x. x\<in> set list}")
apply simp
apply clarsimp
apply (rule_tac x=xa in exI)
apply simp+
apply blast
(*4*)
apply blast

apply clarsimp
apply (subgoal_tac "(\<Union>x\<in>set list. ((\<lambda>(m, s, d, Z). Z) x)) = \<Union>{(\<lambda>(m, s, d, Z). Z) x|x. x\<in> set list}")
apply force
apply force
apply clarsimp

(*1*)
apply (drule_tac x=list in spec)
apply clarsimp
apply (rotate_tac -1, erule impE)
apply (force simp: intersects_def)
apply force
done

(*Now starting from init and doing the whole BC*)
theorem Optimal_BC_existence:
"init\<in>CAN_Nodes C \<Longrightarrow>(\<exists> finalmsgs znlmap. ZNL_BC_msgs C Mid init znlmap finalmsgs []\<and>
 (\<forall> ZN\<in>(HistoryZNL C init finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)))"
apply (subgoal_tac "Connected C ((Entire_Space - CAN_Zones C init)\<union>CAN_Zones C init)")
apply (drule OptimalZNL_existence)
apply simp+
apply (force simp: intersects_def)
apply clarsimp
apply (insert BC_init[of C Mid init])
apply (drule_tac x=" \<lambda> Z N. ZNL" in meta_spec)
apply (erule finish_ZNL_existence)
apply simp+
apply (simp add: Historysimp Message_zone_def Message_dest_def)
apply (insert Connected_Space)
apply (drule Entire_space_nodes)
apply clarsimp
apply (drule Connected_union_singleNode)
apply simp+
done


(*optimal \<Rightarrow> efficiency: no message received twice by the same node -- on an intermediate step *)
theorem efficiency[rule_format]:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap msgs ml;init\<in>CAN_Nodes C\<rbrakk>\<Longrightarrow> (\<forall> ZN\<in>(HistoryZNL C init msgs ml). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN)) \<longrightarrow> (
(\<forall> M M'. M\<in>msgs\<and>M'\<in>msgs\<and>M\<noteq>M'\<longrightarrow>Message_dest M\<noteq>Message_dest M')\<and>
(\<forall> i j. i<length ml\<and>j<length ml \<and>i\<noteq>j\<longrightarrow>Message_dest (ml!i)\<noteq>Message_dest (ml!j))\<and>
(\<forall> M i. M\<in>msgs\<and>i<length ml\<longrightarrow>Message_dest (ml!i)\<noteq>Message_dest M))"
apply (erule broadcast.ZNL_BC_msgs.induct)
 apply (force simp: Message_dest_def)
apply (clarsimp simp: Message_dest_def)
(*1*)
apply (erule impE)
apply (clarsimp simp: Historysimp)
   apply (drule_tac x="(a,b)" in bspec)
   apply ( simp add: Historysimp )
   apply blast
   apply simp
(*1*)
   apply clarsimp
   apply rule
apply clarsimp
apply (erule disjE)
apply clarsimp
apply (erule disjE)
apply clarsimp
(*3*)
apply (rotate_tac -3, drule_tac x=ac in spec)
apply (rotate_tac -1, drule_tac x=ad in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=ba in spec)
apply (rotate_tac -1, drule_tac x=0 in spec)
apply force
(*2*)
apply (erule disjE)
apply (rotate_tac -4, drule_tac x=a in spec)
apply (rotate_tac -1, drule_tac x=aa in spec)
apply (rotate_tac -1, drule_tac x=d in spec)
apply (rotate_tac -1, drule_tac x=b in spec)
apply (rotate_tac -1, drule_tac x=0 in spec)
apply force
(*2*)
apply (drule_tac x=a in spec)
apply (rotate_tac -1, drule_tac x=aa in spec)
apply (rotate_tac -1, drule_tac x=ae in spec)
apply (rotate_tac -1, drule_tac x=b in spec)
apply (rotate_tac -1,drule_tac x=ac in spec)
apply (rotate_tac -1, drule_tac x=ad in spec)
apply (rotate_tac -1, drule_tac x=ae in spec)
apply force
(*1*)
apply rule
apply (clarsimp)
apply (case_tac "i<length ML")
apply (case_tac "j<length ML")
apply (clarsimp simp: nth_append)
apply (rotate_tac 4,drule_tac x="Suc i" in spec)
apply (rotate_tac -1,drule_tac x="Suc j" in spec)
apply force
(*3*)
apply (clarsimp simp: nth_append)
apply (case_tac "(znlmap Z d ! (j - length ML))")
apply (case_tac "ML!i")
apply clarsimp
apply (subgoal_tac "b\<in>NodesInZone C Z")

apply (frule ZNL_dest_outside_zone_Optimal)
apply clarsimp
apply (drule_tac x="(ab,bb)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply force
(*4*)
  apply (rotate_tac -1, drule_tac x="Suc i" in spec)
apply (clarsimp simp: Message_dest_def list_message_nodes_nth intersects_def)
apply (case_tac "ML!i")
apply (drule_tac m=ab and x=bb and y=c and Z=db in ZNL_msgs_Nodes)
apply simp+
apply clarsimp
apply (drule_tac x="(ac,bc)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply (rule Optimal_isValid_ZNL)
   apply force
(*5*)
apply simp+
apply (rule disjI2,rule disjI2, simp add: set_conv_nth)
apply (rule_tac x=" i" in exI)
apply force
apply (drule_tac x=tup in spec)
apply simp+
apply (rotate_tac -1, drule_tac x=0 in spec)
apply (force simp: Message_zone_def)
apply simp+

apply (drule_tac m=aa and x=ba and y=b and Z=da in ZNL_msgs_Nodes)
apply simp+
apply (drule_tac x=ZN in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply (rule Optimal_isValid_ZNL)
   apply force
(*4*)
apply simp+
apply (rule disjI2,rule disjI2, simp add: set_conv_nth)
apply (rule_tac x=" i" in exI)
apply force
(*3*)
apply simp+
apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply (clarsimp simp: Set_of_Optimal_ZNL_def)
   apply (rotate_tac -2, drule_tac x="(j - length ML)" in spec)
   apply clarsimp
   apply (drule sym)
   apply (rotate_tac -1, erule ssubst)
   apply (subgoal_tac "length (znlmap Z d)\<le>length (znlmap Z d)")
   apply (drule_tac Z="(CAN_Zones C b)" and f=fst in intersects_Union_list)
   apply simp+
   apply (erule impE)
apply arith
apply (rule_tac x="(j - length ML)" in exI)
  apply force
  apply simp

(*2*)
apply (case_tac "j<length ML")
apply (clarsimp simp: nth_append)
(* same thing as above exchanging i and j *)
apply (case_tac "(znlmap Z d ! (i - length ML))")
apply (case_tac "ML!j")
apply clarsimp
apply (subgoal_tac "c\<in>NodesInZone C Z")

apply (frule ZNL_dest_outside_zone_Optimal)
apply clarsimp
apply (drule_tac x="(ab,b)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply force
(*4*)
  apply (rotate_tac -1, drule_tac x="Suc j" in spec)
apply (clarsimp simp: Message_dest_def list_message_nodes_nth intersects_def)
apply (drule_tac m=aa and x=ba and y=c and Z=da in ZNL_msgs_Nodes)
apply simp+
apply clarsimp
apply (drule_tac x="(ab,b)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply (rule Optimal_isValid_ZNL)
   apply force
(*5*)
apply simp+
apply (rule disjI2,rule disjI2, simp add: set_conv_nth)
apply (rule_tac x=" j" in exI)
apply force
apply (drule_tac x=tup in spec)
apply simp+
apply (rotate_tac -1, drule_tac x=0 in spec)
apply (force simp: Message_zone_def)
apply simp+

apply (drule_tac m=aa and x=ba and y=c and Z=da in ZNL_msgs_Nodes)
apply simp+
apply (drule_tac x=ZN in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply (rule Optimal_isValid_ZNL)
   apply force
(*4*)
apply simp+
apply (rule disjI2,rule disjI2, simp add: set_conv_nth)
apply (rule_tac x=" j" in exI)
apply force
(*3*)
apply simp+
apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply (clarsimp simp: Set_of_Optimal_ZNL_def)
   apply (rotate_tac -2, drule_tac x="(i - length ML)" in spec)
   apply clarsimp
   apply (drule sym)
   apply (rotate_tac -1, erule ssubst)
   apply (subgoal_tac "length (znlmap Z d)\<le>length (znlmap Z d)")
   apply (drule_tac Z="(CAN_Zones C c)" and f=fst in intersects_Union_list)
   apply simp+
   apply (erule impE)
apply arith
apply (rule_tac x="(i - length ML)" in exI)
  apply force
  apply simp
(*2*)
apply (clarsimp simp: nth_append)
apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply (clarsimp simp: Set_of_Optimal_ZNL_def)
   apply (rotate_tac -2, drule_tac x="(i - length ML)" in spec)
   apply (erule impE)
   apply arith
   apply clarsimp
   apply (rotate_tac -1, drule_tac x="(j - length ML)" in spec)
   apply (erule impE)
   apply arith
   apply (erule impE)
   apply arith
   apply (case_tac "(znlmap Z d ! (i - length ML))")
   apply (case_tac "(znlmap Z d ! (j - length ML))")
   apply force
(*1*)
   apply clarsimp
   apply (case_tac "i<length ML")
apply (clarsimp simp: nth_append)
apply (erule disjE)
apply (rotate_tac 4,drule_tac x=0 in spec)
apply (rotate_tac -1,drule_tac x="Suc i" in spec)
apply clarsimp
apply (rotate_tac -3,drule_tac x=a in spec)
apply (rotate_tac -1,drule_tac x=aa in spec)
apply (rotate_tac -1,drule_tac x="(\<lambda>(m, s, d, Z). d) (ML ! i)" in spec)
apply (rotate_tac -1,drule_tac x=b in spec)
apply (rotate_tac -1,drule_tac x="Suc i" in spec)
apply force
(*1*)
apply (clarsimp simp: nth_append)
apply (case_tac "(znlmap Z d ! (i - length ML))")
apply clarsimp
apply (subgoal_tac "ba\<in>NodesInZone C Z")

apply (erule disjE)

apply (frule ZNL_dest_outside_zone_Optimal)
apply clarsimp
apply (drule_tac x="(ac,bb)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply force
(*3*)
  apply (rotate_tac -1, drule_tac x=0 in spec)
apply (clarsimp simp: Message_dest_def list_message_nodes_nth intersects_def)
apply (drule_tac x=tup in spec)
apply simp+
apply (rotate_tac -1, drule_tac x=0 in spec)
apply (force simp: Message_zone_def)
apply simp+
(*2*)
apply (frule_tac M2="< a| aa, ba, b>" in ZNL_dest_outside_cover_zones_Optimal)
apply clarsimp
apply (drule_tac x="(ac,bb)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply force
   apply force
(*3*)
   apply simp+
apply (clarsimp simp: Message_dest_def list_message_nodes_nth intersects_def Message_zone_def)
(*1*)
apply simp+
apply (drule_tac x="(Z,d)" in bspec)
   apply (simp add: Historysimp Message_zone_def Message_dest_def)
   apply (clarsimp simp: Set_of_Optimal_ZNL_def)
   apply (rotate_tac -2, drule_tac x="(i - length ML)" in spec)
   apply clarsimp
   apply (drule sym)
   apply (rotate_tac -1, erule ssubst)
   apply (subgoal_tac "length (znlmap Z d)\<le>length (znlmap Z d)")
   apply (drule_tac Z="(CAN_Zones C ba)" and f=fst in intersects_Union_list)
   apply simp+
   apply (erule impE)
apply arith
apply (rule_tac x="(i - length ML)" in exI)
  apply force
  apply simp
done

text{*when reaching the final step things are simpler: if the history is optimal THEN we are efficient*}
theorem efficiency_final_step:
"\<lbrakk>ZNL_BC_msgs C Mid init znlmap finalmsgs [];init\<in>CAN_Nodes C; (\<forall> ZN\<in>(HistoryZNL C init finalmsgs []). znlmap (fst ZN) (snd ZN) \<in> Set_of_Optimal_ZNL C (snd ZN) (fst ZN));M \<in>finalmsgs;M'\<in>finalmsgs;M\<noteq>M'\<rbrakk>\<Longrightarrow>Message_dest M\<noteq>Message_dest M'"
apply (frule efficiency)
apply simp
apply simp
apply (elim conjE)
apply (drule_tac x=M in spec)
apply (rotate_tac -1, drule_tac x=M' in spec)
apply (auto simp: Message_dest_def)
done

text{*mixing things together:  from init we reach a final state where BC is efficient *}
theorem efficiency_existence:
"init\<in>CAN_Nodes C \<Longrightarrow>
(\<exists> finalmsgs znlmap. ZNL_BC_msgs C Mid init znlmap finalmsgs []\<and>
(\<forall> M \<in>finalmsgs.\<forall> M'\<in>finalmsgs. M\<noteq>M'\<longrightarrow>Message_dest M\<noteq>Message_dest M'))"
apply (frule Optimal_BC_existence)
apply clarsimp
apply (frule efficiency)
apply simp+
apply (rule_tac x=finalmsgs in exI)
apply rule
apply (rule_tac x=znlmap in exI)
apply simp
apply force
done

text{*
\begin{theorem}[Efficiency]
@{thm efficiency[no_vars]}
\end{theorem}

BC INIT NEW RULE
@{term "ZNL_BC_msgs C Mid init znlmap {} [<Mid|init,init,Entire_Space-CAN_Zones C init>]"}

\begin{theorem}[coverage]
@{thm coverage[no_vars]}
\end{theorem}

\begin{theorem}[Efficiency_existence]
@{thm efficiency_existence[no_vars]}
\end{theorem}

Broadcast

@{typ Broadcast}
\begin{theorem}[Connected_union]
@{thm connected_union_rev[no_vars]}
\end{theorem}

\begin{theorem}[intersects_union]
@{thm intersects_union[no_vars]}
\end{theorem}

\begin{theorem}[finish-ZNL-existence]
@{thm finish_ZNL_existence[no_vars]}
\end{theorem}

*}

end
